from __future__ import absolute_import, unicode_literals

from .compat import abc, parse_qs, urlparse
from .exceptions import DoesNotExist, MultipleObjectsReturned


class Collection(abc.MutableSequence):
    def __init__(self, API, url, params=None):
        if params is None:
            params = {}

        parsed = urlparse(url)
        path, query = parsed.path, parsed.query
        query_params = parse_qs(query)
        query_params = {
            key: value[0] if len(value) == 1 else value
            for key, value in list(query_params.items())
        }

        url = path
        params.update(query_params)

        self.API = API
        self._url = url
        self._params = params

        self._data = None
        self._next_url = None
        self._previous_url = None

    @classmethod
    def from_data(cls, API, response_body):
        result = cls(API, "")
        result._evaluate(response_body)
        return result

    # Evaluate
    @property
    def data(self):
        self._evaluate()
        return self._data

    @property
    def next_url(self):
        self._evaluate()
        return self._next_url

    @property
    def previous_url(self):
        self._evaluate()
        return self._previous_url

    def _evaluate(self, response_body=None):
        if self._data is not None:
            return

        if response_body is None:
            response_body = self.API.request(
                "get", self._url, params=self._params)
        included = {}
        if "included" in response_body:
            included = {
                (item["type"], item["id"]): item for item in response_body["included"]
            }

        self._data = []
        for item in response_body["data"]:
            related = {}
            for (name, relationship) in item.get("relationships", {}).items():
                if relationship is None or "data" not in relationship:
                    continue
                key = (relationship["data"]["type"],
                       relationship["data"]["id"])
                if key in included:
                    related[name] = self.API.new(included[key])
            relationships = item.pop("relationships", {})
            relationships.update(related)
            self._data.append(self.API.new(
                relationships=relationships, **item))

        self._next_url = response_body.get("links", {}).get("next")
        self._previous_url = response_body.get("links", {}).get("previous")

    # Make it look like a list
    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)

    def __setitem__(self, index, value):
        self.data[index] = value

    def __delitem__(self, index):
        del self.data[index]

    def insert(self, index, value):
        self.data.insert(index, value)

    def __repr__(self):
        return repr(self.data)

    def to_dict(self):
        self_url = self._url
        if self._params:
            self_url += "?" + "&".join(
                ("{}={}".format(key, value)
                 for key, value in self._params.items())
            )

        links = {"self": self_url}
        if self.has_next():
            links["next"] = self.next_url()
        else:
            links["next"] = None
        if self.has_previous():
            links["previous"] = self.previous_url()
        else:
            links["previous"] = None

        return {"data": [item.to_dict() for item in self.data], "links": links}

    # Pagination
    def has_next(self):
        return bool(self.next_url)

    def next(self):
        return self.__class__(self.API, self.next_url, self._params)

    def has_previous(self):
        return bool(self.previous_url)

    def previous(self):
        return self.__class__(self.API, self.previous_url, self._params)

    def all_pages(self):
        if self.data:
            yield self
        page = self
        while page.has_next():
            page = page.next()
            yield page

    def all(self):
        for page in self.all_pages():
            for item in page:
                yield item

    # Filters etc
    def filter(self, **filters):
        from .resources import Resource

        params = dict(self._params)

        for key, value in filters.items():
            key = "filter" + "".join(("[{}]".format(part)
                                      for part in key.split("__")))
            if isinstance(value, Resource):
                value = value.id

            params[key] = value

        return self.__class__(self.API, self._url, params)

    def page(self, *args, **kwargs):
        params = dict(self._params)

        if len(args) == 1 and not kwargs:
            params["page"] = args[0]
        elif len(args) == 0 and kwargs:
            for key, value in kwargs.items():
                params["page[{}]".format(key)] = value
        else:
            raise ValueError(
                "Either one positional or keyword arguments " "accepted for pagination"
            )

        return self.__class__(self.API, self._url, params)

    def _param_method(param_name):
        def _method(self, *fields):
            params = dict(self._params)
            params[param_name] = ",".join(fields)
            return self.__class__(self.API, self._url, params)

        return _method

    include = _param_method("include")
    sort = _param_method("sort")
    fields = _param_method("fields")

    def extra(self, **kwargs):
        params = dict(self._params)
        params.update(kwargs)
        return self.__class__(self.API, self._url, params)

    def get(self, **filters):
        if filters:
            qs = self.filter(**filters)
        else:
            qs = self

        if len(qs) == 0:
            raise DoesNotExist()
        if len(qs) > 1:
            raise MultipleObjectsReturned(len(qs))
        return qs[0]



from transifex import native
from transifex.native import tx as _tx
from transifex.native.cache import AbstractCache, MemoryCache
from transifex.native.rendering import (AbstractErrorPolicy,
                                        PseudoTranslationPolicy,
                                        SourceStringErrorPolicy,
                                        SourceStringPolicy)


class DummyErrorPolicy(AbstractErrorPolicy):
    pass


class DummyCache(AbstractCache):
    pass


class TestModuleInit(object):
    """Test __init__.py of root native module."""

    def test_init_uses_given_params(self, reset_tx):
        native.init(
            'mytoken', ['lang1', 'lang2'], cds_host='myhost',
            missing_policy=PseudoTranslationPolicy(),
            error_policy=DummyErrorPolicy(),
            cache=DummyCache(),
        )
        assert _tx.initialized is True
        assert _tx._cds_handler.token == 'mytoken'
        assert _tx._cds_handler.host == 'myhost'
        assert _tx._languages == ['lang1', 'lang2']
        assert isinstance(_tx._missing_policy, PseudoTranslationPolicy)
        assert isinstance(_tx._error_policy, DummyErrorPolicy)
        assert isinstance(_tx._cache, DummyCache)

    def test_initialized_only_once(self, reset_tx):
        # Even if native.init() is called multiple times, only the first one matters
        native.init(
            'mytoken', ['lang1', 'lang2'], cds_host='myhost',
        )
        native.init(
            'another_token',
            ['lang3', 'lang4'],
            'another_host',
            missing_policy=PseudoTranslationPolicy(),
            error_policy=DummyErrorPolicy(),
            cache=DummyCache(),
        )
        assert _tx._cds_handler.token == 'mytoken'
        assert _tx._cds_handler.host == 'myhost'
        assert _tx._languages == ['lang1', 'lang2']
        assert isinstance(_tx._missing_policy, SourceStringPolicy)
        assert isinstance(_tx._error_policy, SourceStringErrorPolicy)
        assert isinstance(_tx._cache, MemoryCache)



# -*- coding: utf-8 -*-
import mock
from django.core.management import call_command
from tests.native.django.test_commands import get_transifex_command
from transifex.common.console import Color

PATH_INVALIDATE_CACHE = ('transifex.native.django.management.utils.push.tx.'
                         'invalidate_cache')


@mock.patch(PATH_INVALIDATE_CACHE)
@mock.patch('transifex.common.console.Color.echo')
def test_invalidate_cache_fail(mock_echo, mock_invalidate_cache):
    mock_invalidate_cache.return_value = 500, {
        'message': 'error message',
        'details': 'error details',
    }

    expected = Color.format(
        '[error]\nCould not invalidate CDS.[end]\n'
        '[high]Status:[end] [warn]{code}[end]\n'
        '[high]Message:[end] [warn]{message}[end]\n'.format(
            code=500,
            message='error message',
        )
    )

    # Make sure it's idempotent
    mock_echo.reset_mock()
    command = get_transifex_command()
    call_command(command, 'invalidate')
    actual = Color.format(mock_echo.call_args_list[1][0][0])
    assert expected == actual


@mock.patch(PATH_INVALIDATE_CACHE)
@mock.patch('transifex.common.console.Color.echo')
def test_invalidate_cache_success(mock_echo, mock_invalidate_cache):
    mock_invalidate_cache.return_value = 200, {
        'data': {
            'count': 5,
        },
    }

    expected = Color.format(
        '[green]\nSuccessfully invalidated CDS cache.[end]\n'
        '[high]Status:[end] [warn]{code}[end]\n'
        '[high]Records invalidated: {count}[end]\n'
        '[high]Note: It might take a few minutes for '
        'fresh content to be available\n'.format(
            code=200,
            count=5,
        )
    )

    # Make sure it's idempotent
    mock_echo.reset_mock()
    command = get_transifex_command()
    call_command(command, 'invalidate')
    actual = Color.format(mock_echo.call_args_list[1][0][0])
    assert expected == actual


@mock.patch(PATH_INVALIDATE_CACHE)
@mock.patch('transifex.common.console.Color.echo')
def test_purge_cache_success(mock_echo, mock_invalidate_cache):
    mock_invalidate_cache.return_value = 200, {
        'data': {
            'count': 5,
        }
    }

    expected = Color.format(
        '[green]\nSuccessfully purged CDS cache.[end]\n'
        '[high]Status:[end] [warn]{code}[end]\n'
        '[high]Records purged: {count}[end]\n'.format(
            code=200,
            count=5,
        )
    )

    # Make sure it's idempotent
    mock_echo.reset_mock()
    command = get_transifex_command()
    call_command(command, 'invalidate', purge=True)
    actual = Color.format(mock_echo.call_args_list[1][0][0])
    assert expected == actual



from __future__ import absolute_import, unicode_literals

import json

# `requests` is somewhat inconsistent with how it raises this exception.
# (https://github.com/psf/requests/issues/5794)
#
# Depending on the environment, the following can happen during
# `response.json`:
#
#   - Python2 and no `simplejson`, a `ValueError` will be raised
#   - Python2 and `simplejson`, a `simplejson.JSONDecodeError` will be raised
#   - Python3 and no `simplejson`, a `json.JSONDecodeError` will be raised
#   - Python3 and `simplejson`, a `simplejson.JSONDecodeError` will be raised
#
# The following wil make sure that catching
# `transifex.api.jsonapi.compat.JSONDecodeError` in a `try: response.json()`
# block will always work
JSONDecodeError = []
try:
    JSONDecodeError.append(json.JSONDecodeError)
except AttributeError:
    pass
try:
    import simplejson

    JSONDecodeError.append(simplejson.JSONDecodeError)
except ImportError:
    pass
if not JSONDecodeError:
    JSONDecodeError = [ValueError]
JSONDecodeError = tuple(JSONDecodeError)

try:
    import collections.abc as abc
except ImportError:
    import collections as abc  # noqa
try:
    from urllib.parse import parse_qs, urlparse
except ImportError:
    from urlparse import parse_qs, urlparse  # noqa



from django.conf import settings

TRANSIFEX_TOKEN = getattr(settings, 'TRANSIFEX_TOKEN', '')
TRANSIFEX_SECRET = getattr(settings, 'TRANSIFEX_SECRET', '')
TRANSIFEX_CDS_HOST = getattr(settings, 'TRANSIFEX_CDS_HOST', None)
TRANSIFEX_MISSING_POLICY = getattr(settings, 'TRANSIFEX_MISSING_POLICY', None)
TRANSIFEX_ERROR_POLICY = getattr(settings, 'TRANSIFEX_ERROR_POLICY', None)
TRANSIFEX_CACHE = getattr(settings, 'TRANSIFEX_CACHE', None)
LANGUAGES = getattr(settings, 'LANGUAGES', [])
SKIP_TRANSLATIONS_SYNC = getattr(settings, 'SKIP_TRANSLATIONS_SYNC', False)
TRANSIFEX_SYNC_INTERVAL = getattr(settings,
                                  'TRANSIFEX_SYNC_INTERVAL',
                                  30*60)
TRANSIFEX_FETCH_ALL_LANGUAGES = getattr(settings,
                                        'TRANSIFEX_FETCH_ALL_LANGUAGES',
                                        False)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json

from transifex.common.utils import (generate_hashed_key, generate_key,
                                    parse_plurals)
from transifex.native.cache import MemoryCache
from transifex.native.cds import CDSHandler
from transifex.native.rendering import (SourceStringErrorPolicy,
                                        SourceStringPolicy, StringRenderer)


class NotInitializedError(Exception):
    """Raised when a method of a TxNative instance is called but the class
    hasn't been initialized.

    Allows for better debugging when developers neglect to call init().
    """
    pass


class TxNative(object):
    """The main class of the framework, responsible for orchestrating all
    behavior."""

    def __init__(self):
        # The class uses an untypical initialization scheme, defining
        # an init() method, instead of initializing inside the constructor
        # This is necessary for allowing it to be initialized by its clients
        # with proper arguments, while at the same time being very easy
        # to import and use a single "global" instance
        self._cache = None
        self._languages = []
        self._error_policy = None
        self._missing_policy = None
        self._cds_handler = None
        self.initialized = False

    def init(
        self, languages, token, secret=None, cds_host=None,
        missing_policy=None, error_policy=None, cache=None,
        fetch_all_langs=False,
    ):
        """Create an instance of the core framework class.

        Also warms up the cache by fetching the translations from the CDS.

        :param list languages: a list of language codes for the languages
            configured in the application
        :param str token: the API token to use for connecting to the CDS
        :param str secret: the additional secret to use for pushing source
            content
        :param str cds_host: an optional host for the Content Delivery Service,
            defaults to the host provided by Transifex
        :param AbstractRenderingPolicy missing_policy: an optional policy
            to use for returning strings when a translation is missing
        :param AbstractErrorPolicy error_policy: an optional policy
            to determine how to handle rendering errors
        :param AbstractCache cache: an optional cache
        :param bool fetch_all_langs: force pull all remote languages
        """
        self._languages = languages
        self._cache = cache or MemoryCache()
        self._missing_policy = missing_policy or SourceStringPolicy()
        self._error_policy = error_policy or SourceStringErrorPolicy()
        self._cds_handler = CDSHandler(
            self._languages, token, secret=secret, host=cds_host,
            fetch_all_langs=fetch_all_langs,
        )
        self.initialized = True

    def translate(
        self, source_string, language_code, is_source=False,
        _context=None, escape=True, params=None, _key=None,
    ):
        """Translate the given string to the provided language.

        :param unicode source_string: the source string to get the translation
            for e.g. 'Order: {num, plural, one {A table} other {{num} tables}}'
        :param str language_code: the language to translate to
        :param bool is_source: a boolean indicating whether `translate`
            is being used for the source language
        :param unicode _context: an optional context that accompanies
            the string
        :param bool escape: if True, the returned string will be HTML-escaped,
            otherwise it won't
        :param dict params: optional parameters to replace any placeholders
            found in the translation string
        :param str _key: an optional key that identifies this string;
            if omitted, the key is generated automatically based on the
            strings itself and its context
        :return: the rendered string
        :rtype: unicode
        """

        if params is None:
            params = {}

        self._check_initialization()

        translation_template = self.get_translation(
            source_string=source_string,
            language_code=language_code,
            _context=_context,
            is_source=is_source,
            _key=_key,
        )

        return self.render_translation(
            translation_template=translation_template,
            params=params,
            source_string=source_string,
            language_code=language_code,
            escape=escape,
        )

    def get_translation(self, source_string, language_code, _context,
                        is_source=False, _key=None):
        """Return the proper translation for the given string
        and language code.

        A translation is a serialized source_string with ICU format
        support, e.g.
        '{num, plural, one {Ένα τραπέζι} other {{num} τραπέζια}}'

        Supports strings in the source language as well, which means
        that if there is a translation available in the cache
        for the source language, it will be used instead of the
        original source_string provided here.
        """
        pluralized, plurals = parse_plurals(source_string)

        if _key is not None:
            # Custom key
            translation_template = self._cache.get(_key, language_code)
        else:
            # Source based key
            _key = generate_key(string=source_string, context=_context)
            translation_template = self._cache.get(_key, language_code)
            if not translation_template:
                # Fallback to hashed based key
                _key = generate_hashed_key(
                    string=source_string, context=_context)
                translation_template = self._cache.get(_key, language_code)

        if (translation_template is not None and pluralized and
                translation_template.startswith('{???')):
            variable_name = source_string[1:source_string.index(',')].strip()
            translation_template = '{{{var}{content}'.format(
                var=variable_name,
                content=translation_template[4:],
            )

        # If rendering the source language and there is no
        # (overridden) translation in the cache, use the original
        # source string
        if is_source and not translation_template:
            translation_template = source_string

        return translation_template

    def render_translation(self, translation_template, params, source_string,
                           language_code, escape=False):
        """Replace the variables in the ICU translation and return the final
        string in the given language.

        If any error occurs during rendering, the error policy is invoked.
        """

        try:
            return StringRenderer.render(
                source_string=source_string,
                string_to_render=translation_template,
                language_code=language_code,
                escape=escape,
                missing_policy=self._missing_policy,
                params=params,
            )
        except Exception:
            return self._error_policy.get(
                source_string=source_string,
                translation=translation_template,
                language_code=language_code,
                escape=escape,
                params=params,
            )

    def fetch_translations(self):
        """Fetch fresh content from the CDS."""
        self._check_initialization()
        self._cache.update(self._cds_handler.fetch_translations())

    def push_source_strings(self, strings, purge=False):
        """Push the given source strings to the CDS.

        :param list strings: a list of SourceString objects
        :param bool purge: True deletes destination source content not included
                           in pushed content.
                           False appends the pushed content to destination
                           source content.
        :return: a tuple containing the status code and the content of the
            response
        :rtype: tuple
        """
        self._check_initialization()
        response = self._cds_handler.push_source_strings(strings, purge)
        return response.status_code, json.loads(response.content)

    def get_push_status(self, job_path):
        """Push the given source strings to the CDS.

        :param str job_path: Job url path
        :return: a tuple containing the status code and the content of the
            response
        :rtype: tuple
        """
        self._check_initialization()
        response = self._cds_handler.get_push_status(job_path)
        return response.status_code, json.loads(response.content)

    def invalidate_cache(self, purge=False):
        """Invalidate CDS cache."""
        self._check_initialization()
        response = self._cds_handler.invalidate_cache(purge)
        return response.status_code, json.loads(response.content)

    def _check_initialization(self):
        """Raise an exception if the class has not been initialized.

        :raise NotInitializedError: if the class hasn't been initialized
        """
        if not self.initialized:
            raise NotInitializedError(
                'TxNative is not initialized, make sure you call init() first.'
            )



import os
from functools import total_ordering

from transifex.common.utils import generate_hashed_key

NO_LOCALE_DIR = object()


@total_ordering
class TranslatableFile(object):
    """Holds information about a localizable file, i.e. a file
    that holds translatable strings."""

    def __init__(self, dirpath, file_name, locale_dir=None):
        self.file = file_name
        self.dirpath = dirpath
        self.locale_dir = locale_dir

    def __repr__(self):
        return "<%s: %s>" % (
            self.__class__.__name__,
            os.sep.join([self.dirpath, self.file]),
        )

    def __eq__(self, other):
        return self.path == other.path

    def __lt__(self, other):
        return self.path < other.path

    @property
    def path(self):
        return os.path.join(self.dirpath, self.file)


class SourceStringCollection(object):
    """Holds SourceString objects in memory."""

    def __init__(self):
        self.strings = {}

    def add(self, source_string):
        """Add a source string to the collection.

        :param SourceString source_string: the object to add
        """
        key = generate_hashed_key(
            string=source_string.string, context=source_string.context
        )
        if key not in self.strings:
            self.strings[key] = source_string
        else:
            self.strings[key].occurrences = source_string.occurrences

    def extend(self, source_strings):
        """Add multiple strings to the collection.

        :param list source_strings: a list of SourceString objects
        """
        if source_strings:
            for string in source_strings:
                self.add(string)

    def update(self, source_strings):
        """Reset the collection so that it only includes the given
         strings.

        :param list source_strings: a list of SourceString objects
        """
        self.strings = {}
        if source_strings is None:
            return

        self.extend(source_strings)



from __future__ import absolute_import, unicode_literals

import json
from copy import deepcopy

import responses
from transifex.api.jsonapi import JsonApi, Resource

from .constants import host


class ATestApi(JsonApi):
    HOST = host


@ATestApi.register
class Foo(Resource):
    TYPE = "foos"


test_api = ATestApi(host=host, auth="test_api_key")


SIMPLE_PAYLOAD = {'type': "foos", 'id': "1", 'attributes': {'hello': "world"}}


def make_simple_assertions(foo):
    assert isinstance(foo, Foo)
    assert foo.TYPE == "foos"
    assert foo.id == "1"
    assert foo.attributes == {'hello': "world"}
    assert foo.relationships == {}
    assert foo.related == {}
    assert foo.hello == "world"


def test_init():
    foo = test_api.Foo(id="1", attributes={'hello': "world"})
    make_simple_assertions(foo)
    foo = test_api.Foo({'data': SIMPLE_PAYLOAD})
    make_simple_assertions(foo)
    foo = test_api.Foo(SIMPLE_PAYLOAD)
    make_simple_assertions(foo)
    foo = test_api.Foo(id="1", hello="world")
    make_simple_assertions(foo)


def test_new():
    foo = test_api.new(type="foos", id="1", attributes={'hello': "world"})
    make_simple_assertions(foo)
    foo = test_api.new({'data': SIMPLE_PAYLOAD})
    make_simple_assertions(foo)
    foo = test_api.new(SIMPLE_PAYLOAD)
    make_simple_assertions(foo)
    foo = test_api.new(type="foos", id="1", hello="world")
    make_simple_assertions(foo)


def test_as_resource():
    foo = test_api.Foo(SIMPLE_PAYLOAD)
    assert (test_api.as_resource(foo).as_resource_identifier() ==
            {'type': "foos", 'id': "1"})
    assert (test_api.as_resource({'data': SIMPLE_PAYLOAD}).
            as_resource_identifier() == {'type': "foos", 'id': "1"})
    assert (test_api.as_resource(SIMPLE_PAYLOAD).
            as_resource_identifier() ==
            {'type': "foos", 'id': "1"})


def test_setattr():
    foo = test_api.Foo(SIMPLE_PAYLOAD)
    foo.hello = "WORLD"
    assert foo.hello == "WORLD"
    assert foo.attributes == {'hello': "WORLD"}


@responses.activate
def test_reload():
    foo = test_api.Foo(SIMPLE_PAYLOAD)

    new_payload = deepcopy(SIMPLE_PAYLOAD)
    new_payload['attributes']['hello'] = "WORLD"
    responses.add(responses.GET, "{}/foos/1".format(host),
                  json={'data': new_payload})

    foo.reload()
    assert foo.hello == "WORLD"
    assert foo.attributes == {'hello': "WORLD"}


@responses.activate
def test_get_one():
    responses.add(responses.GET, "{}/foos/1".format(host),
                  json={'data': SIMPLE_PAYLOAD})

    foo = test_api.Foo.get('1')

    assert len(responses.calls) == 1
    call = responses.calls[0]
    assert call.request.headers['Content-Type'] == "application/vnd.api+json"
    assert call.request.headers['Authorization'] == "Bearer test_api_key"

    make_simple_assertions(foo)


@responses.activate
def test_get_one_with_filters():
    responses.add(responses.GET, "{}/foos?filter[hello]=world".format(host),
                  json={'data': [SIMPLE_PAYLOAD]}, match_querystring=True)

    foo = test_api.Foo.get(hello="world")

    assert len(responses.calls) == 1
    call = responses.calls[0]
    assert call.request.headers['Content-Type'] == "application/vnd.api+json"
    assert call.request.headers['Authorization'] == "Bearer test_api_key"

    make_simple_assertions(foo)


@responses.activate
def test_get_one_with_include():
    responses.add(
        responses.GET,
        "{}/foos/1".format(host),
        json={'data': {'type': "foos",
                       'id': "1",
                       'attributes': {'name': "Foo1"},
                       'relationships': {'sibling': {'data': {'type': "foos",
                                                              'id': "2"}}}},
              'included': [{'type': "foos",
                            'id': "2",
                            'attributes': {'name': "Foo2"}}]},
    )

    foo = test_api.Foo.get('1', include=['sibling'])

    assert len(responses.calls) == 1
    call = responses.calls[0]
    assert call.request.headers['Content-Type'] == "application/vnd.api+json"
    assert call.request.headers['Authorization'] == "Bearer test_api_key"

    assert isinstance(foo, Foo)
    assert foo.TYPE == "foos"
    assert foo.id == "1"
    assert foo.attributes == {'name': "Foo1"}
    assert foo.name == "Foo1"
    assert isinstance(foo.sibling, Foo)
    assert foo.sibling.TYPE == "foos"
    assert foo.sibling.id == "2"
    assert foo.sibling.attributes == {'name': "Foo2"}
    assert foo.sibling.name == "Foo2"


@responses.activate
def test_save_existing():
    foo = test_api.Foo(SIMPLE_PAYLOAD)

    new_payload = deepcopy(SIMPLE_PAYLOAD)
    new_payload['attributes']['hello'] = "WORLD"
    responses.add(responses.PATCH, "{}/foos/1".format(host),
                  json={'data': new_payload})

    foo.hello = "WORLD"
    foo.save()

    assert len(responses.calls) == 1
    call = responses.calls[0]
    assert json.loads(call.request.body.decode()) == {'data': new_payload}

    foo.hello = "something else"
    foo.save(hello="WORLD")

    assert len(responses.calls) == 2
    call = responses.calls[1]
    assert json.loads(call.request.body.decode()) == {'data': new_payload}


@responses.activate
def test_save_new():
    new_payload = deepcopy(SIMPLE_PAYLOAD)
    new_payload['attributes']['created'] = "NOW!!!"
    responses.add(responses.POST, "{}/foos".format(host),
                  json={'data': new_payload})

    foo = test_api.Foo(attributes={'hello': "world"})
    foo.save()

    assert foo.id == "1"
    assert foo.created == "NOW!!!"
    assert foo.attributes == {'hello': "world", 'created': "NOW!!!"}


@responses.activate
def test_create():
    new_payload = deepcopy(SIMPLE_PAYLOAD)
    new_payload['attributes']['created'] = "NOW!!!"
    responses.add(responses.POST, "{}/foos".format(host),
                  json={'data': new_payload})

    foo = test_api.Foo.create(attributes={'hello': "world"})

    assert foo.created == "NOW!!!"
    assert foo.attributes == {'hello': "world", 'created': "NOW!!!"}


@responses.activate
def test_create_with_id():
    new_payload = deepcopy(SIMPLE_PAYLOAD)
    new_payload['attributes']['created'] = "NOW!!!"
    responses.add(responses.POST, "{}/foos".format(host),
                  json={'data': new_payload})

    foo = test_api.Foo.create(id="2", attributes={'hello': "world"})

    assert len(responses.calls) == 1
    assert json.loads(responses.calls[0].request.body.decode()) == {'data': {
        'type': "foos",
        'id': "2",
        'attributes': {'hello': "world"},
    }}
    assert foo.created == "NOW!!!"
    assert foo.attributes == {'hello': "world", 'created': "NOW!!!"}
    assert foo.id == "1"


@responses.activate
def test_delete():
    responses.add(responses.DELETE, "{}/foos/1".format(host))

    foo = test_api.Foo(SIMPLE_PAYLOAD)
    foo.delete()

    assert len(responses.calls) == 1
    assert foo.id is None


def test_eq():
    foo = test_api.Foo(SIMPLE_PAYLOAD)
    assert foo == {'type': "foos", 'id': "1"}
    assert {'type': "foos", 'id': "1"} == foo
    assert foo == {'data': {'type': "foos", 'id': "1"}}
    assert {'data': {'type': "foos", 'id': "1"}} == foo
    assert foo == test_api.Foo(id="1")
    assert test_api.Foo(id="1") == foo


def test_as_resource_identifier():
    foo = test_api.Foo(SIMPLE_PAYLOAD)
    assert foo.as_resource_identifier() == {'type': "foos", 'id': "1"}


def test_as_relationship():
    foo = test_api.Foo(SIMPLE_PAYLOAD)
    assert foo.as_relationship() == {'data': {'type': "foos", 'id': "1"}}



from __future__ import unicode_literals

import ast

import asttokens
from transifex.common.console import Color
from transifex.common.strings import (VAR_FORMAT, alt_quote,
                                      printf_to_format_style)
from transifex.native.consts import KEY_CONTEXT
from transifex.native.parsing import (ENCODING_PATTERN, CallDetectionVisitor,
                                      get_func_parts)
from transifex.native.tools.migrations.mark import MARK_ARGUMENT_FIXME
from transifex.native.tools.migrations.models import (Confidence,
                                                      FileMigration,
                                                      StringMigration)

# Simple string
GETTEXT = 'get'
UGETTEXT = 'uget'

# Pluralized string
NGETTEXT = 'nget'
UNGETTEXT = 'unget'

# String with context
PGETTEXT = 'pget'

# Pluralized string with context
NPGETTEXT = 'npget'

LAZY_SUFFIX = '_lazy'

# Lazy variants
GETTEXT_LAZY = GETTEXT + LAZY_SUFFIX
UGETTEXT_LAZY = UGETTEXT + LAZY_SUFFIX
NGETTEXT_LAZY = NGETTEXT + LAZY_SUFFIX
UNGETTEXT_LAZY = UNGETTEXT + LAZY_SUFFIX
PGETTEXT_LAZY = PGETTEXT + LAZY_SUFFIX
NPGETTEXT_LAZY = NPGETTEXT + LAZY_SUFFIX

LAZY_MAPPING = {
    GETTEXT: GETTEXT_LAZY,
    UGETTEXT: UGETTEXT_LAZY,
    NGETTEXT: NGETTEXT_LAZY,
    UNGETTEXT: UNGETTEXT_LAZY,
    PGETTEXT: PGETTEXT_LAZY,
    NPGETTEXT: NPGETTEXT_LAZY,
}

KEYWORD_STRING = 'string'
KEYWORD_CONTEXT = 'context'
KEYWORD_ONE = 'one'
KEYWORD_OTHER = 'other'
KEYWORD_CNT = 'cnt'

ICU_PLURAL_COUNTER = 'cnt'


class GettextMethods(object):
    """Configuration class that holds paths to gettext wrapper methods
    as well as expected arguments.

    For each 3rd-party framework that uses gettext for localization, we need
    to pass different paths and arguments, so that calls to these methods
    are properly migrated.
    """

    def __init__(self, **kwargs):
        """Constructor.

        Usage:
        >>> GettextMethods(
        >>>     uget=(
        >>>         'path.to.ugettext',
        >>>         [('message', KEYWORD_STRING)],
        >>>     ),
        >>>     nget= (
        >>>         'path.to.ngettext',
        >>>         [
        >>>             ('singular', KEYWORD_ONE),
        >>>             ('plural', KEYWORD_OTHER),
        >>>             ('number', KEYWORD_CNT),
        >>>         ]
        >>>     ),
        >>> )
        """
        self.methods = kwargs

    def gettext_type_from_path(self, full_path):
        """Get the type of gettext method the given import path corresponds to.

        Supported types are UGET, UNGET, etc.

        Usage:
        >>> gettext_type_from_path('django.utils.translation.ngettext')
        'nget'

        :param str full_path: an import path
        :return: a string type
        :rtype: str
        """
        for gettext_type, params in self.methods.items():
            if params[0] == full_path:
                return gettext_type
        return None

    def tx_native_details_from_type(self, gettext_type):
        """Return details on how to structure a Native call of the given
        gettext type.

        Usage:
        >>> tx_native_details_from_type(UGETTEXT)
        <<< (
        <<<     't',
        <<<     (
        <<<         'django.utils.translation.ugettext',
        <<<         [('message', 'string')],
        <<<     )
        <<< )

        :param str gettext_type: e.g. UGET, UNGET, etc
        :return: a tuple like (<new_func_name>, [...])
        :rtype: tuple
        """
        params = self.methods.get(gettext_type)
        if not params:
            raise ValueError(
                'Unregistered or invalid gettext type "{}"'.format(
                    gettext_type
                )
            )
        method_name = params[0]
        native_method_name = (
            'lazyt' if method_name.endswith(LAZY_SUFFIX) else 't'
        )
        return native_method_name, params[1]

    @property
    def all(self):
        """Return a list of the parameters of all registered methods."""
        return self.methods.values()


class GettextMigrationBuilder(object):
    """Parses Python code and creates file migrations for each template.

    A migration is an object that describes the changes that need to be made
    in order to change a Python file that uses the gettext syntax
    to a file that uses the Transifex Native syntax.

    This class is stateful, but it can be reused. The reason it is created
    this way is for optimization, i.e. for not having to create new instances
    of the class for each migrated file.
    """

    def __init__(self, methods, import_statement):
        """Constructor.

        :param GettextMethods methods: the configuration to use
        :param str import_statement: the import statement to use
            in all migrated files
        """
        self._reset()
        self.methods = methods
        self.import_statement = import_statement
        self.transformer = Transformer(self.methods, import_statement)

    def _reset(self):
        """Reset some state parameters, so that a new migration can be created.

        The reason this method exists is for optimization, i.e. not having
        to create new instances of the class for each file.
        """
        # Sometimes, like when a comment is available, we cannot parse
        # a translatable string in one go, so we need to create
        # a StringMigration object and amend it as soon as we have all its
        # information.
        self._current_string_migration = None

    def build_migration(self, src, filename=None):
        """Create a migration for a Django template file to Transifex
        Native syntax.

        The returned object contains every change separately, so that
        it can be reviewed string by string.

        :param unicode src: the whole Django template
        :param str filename: the filename of the original template
        :return: a FileMigration instance
        :rtype: FileMigration
        """
        self._reset()
        file_migration = self.transformer.transform(src, filename)
        return file_migration


class Transformer(object):
    """Transforms source strings inside Python files from gettext syntax
    to Transifex Native syntax.

    It also keeps stats about any errors that have occurred.

    It allows clients to register custom module/function paths in order
    to support gettext module/function wrappers that exist in framework
    implementations that use native, such as a Django or a Flask SDK.
    """

    def __init__(self, methods, import_statement):
        """Constructor.

        :param GettextMethods methods: the gettext methods to transform
        :param str import_statement: the import statement to use
            in all migrated files
        """
        self.errors = []
        self._methods = methods
        if not import_statement.endswith('\n'):
            import_statement = '{}\n'.format(import_statement)
        self.import_statement = import_statement
        self._functions = []
        self.register_functions(*self._methods.all)

    def register_functions(self, *func_details_list):
        """Register a custom function to be detected during extraction.

        Each arg in `func_paths` must be a tuple, with its first item
        a string representing the full path of the function
        (like 'module.deeper_module.func_name') and its second item
        a list of tuples representing the items.

        Example:
        (
            'django.utils.translation.ugettext',
            [('message', 'string')],
        )
        """
        for func_path, argument_pairs in func_details_list:
            nodes = func_path.split('.')
            if len(nodes) < 1:
                raise ValueError(
                    'Function path must contain zero or more modules and '
                    'a function, e.g. "my_module.translate"'
                )

            self._functions.append(
                {
                    'modules': '.'.join(nodes[:-1]),
                    'function': nodes[-1],
                    'arguments': argument_pairs,
                }
            )

    def transform(self, src, filename=None):
        """Parse the given Python file string and extract translatable content.

        :param unicode src: a chunk of Python code
        :param str filename: the filename of the code, i.e. the filename it
            came from
        :return: a list of SourceString objects
        :rtype: list
        """
        # Replace utf-8 magic comment, to avoid getting a
        # "SyntaxError: encoding declaration in Unicode string"
        src = ENCODING_PATTERN.sub('# ', src)
        try:
            tree = ast.parse(src)
            visitor = CallDetectionVisitor(
                [
                    (x['modules'], x['function'])
                    for x in self._functions
                ]
            )
            visitor.visit(tree)

        except Exception as e:
            Color.echo(
                '[error]Error while parsing content '
                'of [file]{}[end]: {}'.format(filename, e)
            )
            # Store an exception for this particular file
            self.errors.append((filename, e))
            return None

        else:
            attree = asttokens.ASTTokens(src, tree=tree)
            file_migration = FileMigration(filename, src)
            last_migrated_char = 0

            def add_in_between(txt_range):
                """Add the simple text found between the last migrated node
                and the current node.

                Ensures that the text that should remain untransformed
                is included in the final file.
                """
                if txt_range[0] > last_migrated_char:
                    try:
                        original_str = src[last_migrated_char:txt_range[0]]
                        file_migration.add_string(
                            StringMigration(
                                original=original_str,
                                new=original_str,
                                confidence=Confidence.HIGH,
                            )
                        )
                    except Exception as e:
                        Color.echo(
                            '[error]Error while adding in-between content '
                            'in [file]{}[end]. last_migrated_char={}.'
                            'Error: {}'.format(
                                filename, last_migrated_char, e)
                        )
                        raise

            # Create a map with the text range for each node to migrate
            # We need this in order to sort the nodes. This way, we can
            # support the migration of imports that appear after function
            # calls (e.g. locally)
            text_ranges = {}
            to_migrate = (
                [x.node for x in visitor.imports] +
                visitor.function_calls
            )
            for node in to_migrate:
                text_ranges[node] = attree.get_text_range(node)

            # Remove duplicates
            to_migrate = sorted(set(to_migrate),
                                key=lambda n: text_ranges[n][0])

            import_added = False

            # Create a migration for adding the import statement
            # of Native. At this moment we don't know if it will need
            # to include t, lazyt or both, but we'll update the instance
            # after all nodes have been processed
            native_import_string_migration = StringMigration(
                '', ''
            )
            native_functions = set()  # will store 't'/'lazyt' if found later

            for node in to_migrate:
                text_range = text_ranges[node]
                add_in_between(text_range)

                try:
                    original = src[text_range[0]:text_range[1]]
                    new = original
                    confidence = Confidence.HIGH

                    # Migrate ImportFrom nodes. Leave Import nodes intact,
                    # as they may have been added in the code for uses other
                    # than gettext calls
                    if isinstance(node, ast.ImportFrom):
                        try:
                            #
                            if not import_added:
                                file_migration.add_string(
                                    native_import_string_migration
                                )
                            new, item_native_functions = self._transform_import(
                                visitor, node)
                            confidence = Confidence.HIGH
                            import_added = True
                            native_functions.update(item_native_functions)

                            # If the whole import statement was about gettext
                            # functions, a new empty line will have been
                            # added to the file. In that case,
                            # this will also remove the empty line completely
                            if new == '' and \
                                    node.first_token.line == original + '\n':
                                text_range = (text_range[0], text_range[1] + 1)
                                original = original + '\n'
                        except Exception as e:
                            raise Exception(
                                'Error while migrating import'
                                '\n{}'.format(str(e))
                            )

                    # Migrate function calls
                    elif isinstance(node, ast.Call):
                        try:
                            new, confidence = self._transform_call(
                                node, visitor, attree)
                        except Exception as e:
                            raise Exception(
                                'Error while migrating call'
                                '\n{}'.format(str(e))
                            )

                        # If this function call was part of a % operation
                        # make sure the right part is also removed
                        modulo_node = visitor.modulos.get(node)
                        if modulo_node:
                            # Override the text range with that of the
                            # modulo operation. This includes the function
                            # call as well
                            text_range = attree.get_text_range(modulo_node)
                            original = src[text_range[0]:text_range[1]]

                    file_migration.add_string(
                        StringMigration(
                            original=original,
                            new=new,
                            confidence=confidence,
                        )
                    )
                    last_migrated_char = text_range[1]

                except Exception as e:
                    Color.echo(
                        '[error]Error while transforming content '
                        'of [file]{}[end]: {}'.format(filename, e)
                    )
                    Color.echo('Original content:\n{}'.format(original))
                    # Store an exception for this particular file
                    self.errors.append((filename, e))
                    return None

            # Add the rest of the file (from the last migrated node
            # to the end)
            if last_migrated_char < len(src):
                original = src[last_migrated_char:]
                file_migration.add_string(
                    StringMigration(original=original, new=original)
                )

            # Update the Native import statement with the proper functions
            if native_functions:
                native_import_string_migration.update(
                    extra_original='',
                    extra_new=self.import_statement.format(
                        ', '.join(sorted(native_functions))
                    )
                )

            return file_migration

    def _transform_import(self, visitor, import_node):
        """Make sure any imports that are not related to gettext
        translations are not removed in the migrated string.

        If `migrated_import_string` is empty, it doesn't do anything.
        Otherwise, it searches to see if the particular ImportFrom node
        included any other imports that are not registered gettext methods
        and should be kept in the migration.

        For example, this:
            from django.utils.translation import ugettext as _, foo, bar as _bar
            from django.utils import translation as _trans
            import django
        will be transformed to this (in 3 consecutive _transform_import calls):
            from transifex.native.django import t
            from django.utils.translation import foo, bar as _bar
            from django.utils import translation as _trans

        :param CallDetectionVisitor visitor: the visitor object
        :param ast.Node import_node: the Import or ImportFrom node
        :return: a new string to be used as the "new" migration string
        :rtype: unicode
        """
        if isinstance(import_node, ast.Import):
            return '', []

        full_registered_paths = ['{}.{}'.format(x['modules'], x['function'])
                                 for x in self._functions]

        imports_per_node = visitor.imports_per_node.get(import_node)
        if not imports_per_node:
            return '', []

        imports = visitor.imports_per_node[import_node]['imports']
        module = visitor.imports_per_node[import_node]['module']
        all_import_units = [
            (unit, '{}.{}'.format(module, unit[0]))
            for unit in imports
        ]
        imports_to_keep = []
        for unit, full_path in all_import_units:
            if full_path not in full_registered_paths:
                imports_to_keep.append(unit)

        method_names = [
            self._methods.gettext_type_from_path(x[1])
            for x in all_import_units
        ]
        function_names = [
            'lazyt' if method_name.endswith(LAZY_SUFFIX) else 't'
            for method_name in method_names
            if method_name is not None
        ]
        if not imports_to_keep:
            return '', function_names

        new_string = 'from {module} import {units}'.format(
            module=module,
            units=', '.join([
                '{} as {}'.format(
                    unit[0], unit[1]
                )
                if unit[1] else unit[0]
                for unit in imports_to_keep
            ])
        )
        return new_string, function_names

    def _transform_call(self, func_call_node, visitor, attree):
        """Transform a function call to Transifex Native format.

        :param ast.Node func_call_node: the source node
        :param CallDetectionVisitor visitor: the visitor object
        :return: a tuple with the new serialized code and the confidence
            level of the migration
        :rtype: Tuple[unicode, int]
        """
        module_path, func_name = get_func_parts(func_call_node)

        for import_obj in visitor.imports:
            if module_path != import_obj.module \
                    or func_name != import_obj.function:
                continue

            for import_unit in import_obj.node.names:
                name = import_unit.name
                asname = import_unit.asname or name
                if func_name == asname:
                    break

            try:
                module = import_obj.node.module
            except AttributeError:
                module = None

            if not module_path:
                full_path = '{module}.{name}'.format(
                    module=module, name=name)
            else:
                full_path = '{module}{path}.{func}'.format(
                    module=('{}.'.format(module) if module else ''),
                    path=module_path.replace(
                        asname,
                        name,
                        1,
                    ),
                    func=func_name
                )

            if not full_path:
                name = import_obj.node.names[0].name
                asname = import_obj.node.names[0].asname or name

            gettext_type = self._methods.gettext_type_from_path(full_path)
            if gettext_type is None:
                continue

            new_func_name, args = self._methods.tx_native_details_from_type(
                gettext_type
            )
            new_arguments_serialized, confidence = self._serialize_arguments(
                func_call_node, gettext_type, args
            )

            # Update the source string to match TX Native syntax
            # and replace it in the arguments list
            new_source_string, placeholders1 = printf_to_format_style(
                new_arguments_serialized[0]
            )

            new_arguments_serialized[0] = new_source_string
            quote = new_source_string[0]

            modulo_node = visitor.modulos.get(func_call_node)
            interpolated_params, placeholders2, modulo_confidence = \
                self._parse_modulo_operation(modulo_node, quote, attree)
            if modulo_confidence == Confidence.LOW:
                confidence = Confidence.LOW

            # If they don't match, it means that we haven't properly
            # detected and migrated the values
            if placeholders1 != placeholders2:
                confidence = Confidence.LOW

            new_call = '{func_name}({args})'.format(
                func_name=new_func_name,
                args=', '.join(new_arguments_serialized + interpolated_params)
            )
            return new_call, confidence

        # Normally we shouldn't reach this point
        return None, Confidence.LOW

    def _parse_modulo_operation(self, modulo_node, quote, attree):
        """Extract all parameters from the given modulo node.

        The returned parameters can be used as is on a serialized
        t() function call.

        :param ast.Node modulo_node: the node to parse
        :param str quote: the quote to use for wrapping strings, i.e. " or '
        :param asttokens.ASTTokens attree: the parsed tree
        :return: a tuple structured like:
            (
                ['var1=var1', 'var2="value"', 'username=username'],
                ['var1', 'var2', 'username'],
                Confidence.HIGH
            )
        :rtype: Tuple[list, list, int]
        """
        if modulo_node is None:
            return [], [], Confidence.HIGH

        def var(num):
            """Generate a name for a variable with unknown name."""
            return VAR_FORMAT.format(cnt=str(num))

        def extract_param(_name, _element):
            """Get a parameter assignment for use with t() syntax,
            depending on the type of the given element"""
            if isinstance(_element, ast.Name):
                # variables: <name>=<value>
                return '{}={}'.format(_name, _element.id)
            elif isinstance(_element, ast.Str):
                # literals: <name>="<value>"
                new_quote = alt_quote(quote, _element.s)
                return ''.join([_name, '=', new_quote, _element.s, new_quote])
            elif isinstance(_element, ast.BinOp):
                # operations: <name>=(<left> <op> <right>)
                return '{}=({})'.format(_name, attree.get_text(_element))
            else:
                # everything else: <name>=<value>
                return '{}={}'.format(_name, attree.get_text(_element))

        def confidence(_element):
            """Return the proper confidence for the given node.

            It returns low confidence if the text includes \\' or \\"
            as these are not being properly handled by alt_quote().
            """
            if isinstance(_element, ast.Name):
                _string = _element.id
            elif isinstance(_element, ast.Str):
                _string = _element.s
            else:
                _string = attree.get_text(_element)
            if r"\\'" in _string or r'\\"' in _string:
                return Confidence.LOW
            return Confidence.HIGH

        right_op = modulo_node.right
        placeholders = []

        # e.g. % something
        #      % 'something'
        if isinstance(right_op, (ast.Name, ast.Str)):
            param = var(1)
            return [extract_param(param, right_op)], [param], confidence(
                right_op)

        # e.g. % (something, "literal")
        if isinstance(right_op, (ast.List, ast.Tuple)):
            cnt = 1
            params = []
            confidences = []
            for element in right_op.elts:
                name = var(cnt)
                params.append(extract_param(name, element))
                placeholders.append(name)
                cnt += 1
                confidences.append(confidence(element))

            final_confidence = (
                Confidence.LOW if Confidence.LOW in confidences
                else Confidence.HIGH
            )
            return params, placeholders, final_confidence

        # e.g. % {'foo': foo, 'bar': 'bar'}
        if isinstance(right_op, ast.Dict):
            params = []
            confidences = []
            for key, element in zip(right_op.keys, right_op.values):
                name = key.s
                params.append(extract_param(name, element))
                placeholders.append(name)
                confidences.append(confidence(element))

            final_confidence = (
                Confidence.LOW if Confidence.LOW in confidences
                else Confidence.HIGH
            )
            return params, placeholders, final_confidence

        # e.g. % dict(foo=foo, bar='bar')
        # NOTE: We could instead parse all values of the dictionary
        # but we chose not to for now
        if (isinstance(right_op, ast.Call)
                and hasattr(right_op.func, 'id')
                and right_op.func.id == 'dict'):
            string = attree.get_text(right_op)
            new_quote = alt_quote(quote, string)
            value = ''.join(
                [MARK_ARGUMENT_FIXME, '=', new_quote, string, new_quote]
            )
            return [value], [MARK_ARGUMENT_FIXME], Confidence.LOW

        # e.g. % (3 * 15)
        if isinstance(right_op, ast.BinOp):
            name = var(1)
            # Wrap in parentheses to make sure it's valid
            return [extract_param(name, right_op)], [name], Confidence.HIGH

        # In any other case, get the node text and use it as is
        # in a variable
        name = var(1)
        return [extract_param(name, right_op)], [name], Confidence.HIGH

    def _serialize_arguments(self, func_call_node, gettext_type, expected_args):
        """Return arguments as a serialized string for the given function call,
        as transformed to fit the syntax of Transifex Native.

        Supports parameters passed in all possible ways:
          - gettext('text')
          - gettext(*['text'])
          - gettext(message='text')
          - gettext(**{'message': 'text'})

        Example return:
        (
            ['This is the {foo} source string', '_context="A context"'],
            Confidence.HIGH
        )

        :param ast.Call func_call_node: the function call node
        :param str gettext_type: the type of method, e.g. NGETTEXT
        :param list expected_args: a list of tuples
        :return: a tuple that contains:
            - a list with strings of all arguments serialized, ready to be used
              in a Native call
            - a Confidence value, based on the parsed arguments
        :rtype: Tuple[list, int]
        """
        def replace_quotes(_string, existing_quote):
            """Return new string with alternate quotes, if necessary.

            Assumes default quotes are double quotes (").
            :rtype: unicode
            """
            if not existing_quote or existing_quote == '"':
                return _string
            return _string.replace('"', "'")

        # Make a copy of all containers, we'll delete items as we go
        args = list(func_call_node.args)
        starargs = (
            list(func_call_node.starargs.elts)
            if hasattr(func_call_node, 'starargs') and func_call_node.starargs
            else None
        )
        keywords = list(func_call_node.keywords)
        kwargs = (
            list(func_call_node.kwargs.keywords)
            if hasattr(func_call_node, 'kwargs') and func_call_node.kwargs
            else {}
        )

        new_arguments = {}
        for arg_name, arg_type in expected_args:
            # Start with normal arguments, e.g. func('value')
            node = args[0] if args else None
            if node:
                args = args[1:]
                new_arguments[arg_type] = (render_keyword(node),
                                           original_quote(node))
                continue

            # If not found, try starargs, e.g. func(*['value'])
            node = starargs[0] if starargs else None
            if node:
                starargs = starargs[1:]
                new_arguments[arg_type] = (render_keyword(node),
                                           original_quote(node))
                continue

            # If not found, try keywords, e.g. func(message='value')
            found = False
            for keyword in keywords:
                # This only happens on Python 3 for calls like:
                # func(**dict(message='value')). In this case we switch
                # the current `keyword` # with that found inside **dict(...)
                # The `if` statement that follows will take care of the actual
                # parsing and rendering
                if keyword.arg is None and isinstance(keyword.value, ast.Call):
                    inner_func_call = keyword.value
                    if inner_func_call.func.id != 'dict':
                        continue
                    for k in inner_func_call.keywords:
                        if k.arg == arg_name:
                            keyword = k
                            break

                if keyword.arg == arg_name:
                    new_arguments[arg_type] = (render_keyword(keyword.value),
                                               original_quote(keyword))
                    if keyword in keywords:
                        keywords.remove(keyword)
                    found = True
                    break
            if found:
                continue

            # Final resort, try kwargs, e.g. func(**{'message': 'value'})
            for keyword in kwargs:
                if keyword.arg == arg_name:
                    new_arguments[arg_type] = (render_keyword(keyword.value),
                                               original_quote(keyword))
                    kwargs.remove(keyword)
                    break

        # Simple string
        if gettext_type in (GETTEXT, UGETTEXT, GETTEXT_LAZY, UGETTEXT_LAZY):
            string, quote = new_arguments[KEYWORD_STRING]
            return [replace_quotes('"{}"', quote).format(string)], \
                Confidence.HIGH

        # Pluralized string
        if gettext_type in (NGETTEXT, UNGETTEXT, NPGETTEXT, NGETTEXT_LAZY, UNGETTEXT_LAZY, NPGETTEXT_LAZY):
            one, quote_one = new_arguments[KEYWORD_ONE]
            other, quote_other = new_arguments[KEYWORD_OTHER]

            confidence = Confidence.HIGH
            if quote_one != quote_other and quote_one in other:
                confidence = Confidence.LOW
                other = other.replace(quote_one, r'\{}'.format(quote_one))

            items = [
                replace_quotes('"{', quote_one), ICU_PLURAL_COUNTER, ', ',
                'one {', one, '} ',
                'other {', other, '}',
                replace_quotes('}"', quote_one)  # end of ICU string
            ]
            # Also includes context
            if gettext_type == NPGETTEXT:
                context, quote_context = new_arguments[KEYWORD_CONTEXT]
                items.extend((
                    ', ', KEY_CONTEXT, replace_quotes('="', quote_context),
                    context, replace_quotes('"', quote_context),
                ))
            items.extend((
                ', ', ICU_PLURAL_COUNTER, '=', new_arguments[KEYWORD_CNT][0]
            ))

            return [''.join([str(x) for x in items])], confidence

        # String with context
        if gettext_type in (PGETTEXT, PGETTEXT_LAZY):
            string, string_quote = new_arguments[KEYWORD_STRING]
            context, quote_context = new_arguments[KEYWORD_CONTEXT]
            string = replace_quotes('"{}"', string_quote).format(string)
            context = replace_quotes('_context="{}"', quote_context).format(
                context
            )
            return [string, context], Confidence.HIGH

        raise ValueError(
            'Invalid gettext type: {}. Cannot serialize arguments'.format(
                gettext_type
            )
        )


def render_keyword(keyword):
    """Render the given keyword to a proper value.

    Processes Keyword objects and returns its value, using the correct property
    depending on its type.

    :param Keyword keyword: the keyword object
    :return: the keyword value (string, number, etc)
    :rtype: object
    """
    if isinstance(keyword, ast.Str):
        return keyword.s
    if isinstance(keyword, ast.Num):
        return keyword.n
    if isinstance(keyword, ast.Name):
        return keyword.id
    return ''


def original_quote(node):
    """Return the quote (' or ") found in the first character of the
    given node.

    :param ast.Node node: the node to check
    :return: a 1-char string, ' or "
    :rtype: str
    """
    quote = node.last_token.string[0]
    return quote if quote in ('"', "'") else None



from __future__ import unicode_literals

# A single quote that is not preceded by \
# i.e. matches "This is 'something" but not "This is \' something"
import re

from transifex.common._compat import PY3, binary_type, text_type

VAR_FORMAT = 'variable_{cnt}'

RE_SINGLE_QUOTE = r"(?<!\\)\'"
# A double quote that is not preceded by \
# i.e. matches 'This is "something' but not 'This is \" something'
RE_DOUBLE_QUOTE = r'(?<!\\)\"'


def printf_to_format_style(string):
    """Transform any %s-style placeholders in the given string to
    str.format() placeholders.

    Usage:
    >>> printf_to_format_style('This is %s & %s')
    <<< ('This is {variable_1} & {variable_2}', '['variable_1', 'variable_2']')

    >>> printf_to_format_style('This is %(foo)s and %(bar)s')
    <<< ('This is {foo} and {bar}', ['foo', 'bar'])

    :param unicode string: the source string
    :return: a tuple, containing the new string (with str.format() syntax)
        and a list of the names of all variables
    :rtype: tuple
    """
    obj = {'cnt': 1, 'variables': []}  # Python 2 nonlocal workaround

    def replace_named(match):
        """Given a regex match like '%(foo)s' return '{foo}'.

        Also stores the found variable inside `obj`.

        :param match: a regex match object
        :return: a new string using str.format() placeholder syntax
        :rtype: str
        """
        # [2:-2] means from '%(foo)s' -> get 'foo'
        var = match.group(0)[2:-2]
        new = '{{{}}}'.format(var)
        obj['variables'].append(var)
        return new

    pattern = re.compile(r'(%\(\w+\)s)')
    new_string, total = re.subn(
        pattern,
        replace_named,
        string
    )

    def replace_unnamed(match):
        """Given a regex match like '%s' return '{variable_N}' where N
        is an auto-increase integer.

        Also stores the found variable inside `obj`.

        :param match: a regex match object
        :return: a new string using str.format() placeholder syntax
        :rtype: str
        """
        var = VAR_FORMAT.format(cnt=str(obj['cnt']))
        new = '{{{}}}'.format(var)
        obj['variables'].append(var)
        obj['cnt'] += 1
        return new

    pattern = re.compile('(%s)')
    new_string, total = re.subn(pattern, replace_unnamed, new_string)

    return new_string, obj['variables']


def alt_quote(quote, string):
    """Return the proper quote character to use for wrapping the given string.

    Takes into account any identical quotes that are included in the string
    and, if found, it returns the alternate quote type, to avoid breaking the
    code if the string is wrapped in quotes.

    Usage:
    >>> alt_quote('"', 'This is a string')
    '"' (double quote)
    >>> alt_quote('"', 'This is a " string')
    "'" (single quote)
    >>> alt_quote('"', r'This is a \" string')
    '"' (double quote)

    :param unicode quote: either ' or ", the preferred quote to use for
        wrapping `string`
    :param unicode string: the string that will be wrapped in quotes
    :return: a new string
    :rtype: unicode
    """
    alternate = '"' if quote == "'" else "'"
    pattern = RE_SINGLE_QUOTE if quote == "'" else RE_DOUBLE_QUOTE
    pattern_alt = RE_SINGLE_QUOTE if quote == '"' else RE_DOUBLE_QUOTE
    if re.search(pattern, string) and not re.search(pattern_alt, string):
        return alternate
    return quote


class LazyString(object):
    """Can be used instead of a string instance when delayed evaluation
    is desired.

    Upon instantiation, the caller needs to provide a function along
    with any parameters, that will be used when the string value will be
    evaluated.

    Lazy evaluation is achieved through Pythons magic methods `__str__`
    and `__unicode__`.

    Usage:
    In the following example, the value of 'foo' and 'bar' is not available
    # when the string is declared (mapping is empty). However, mapping is only
    # accessed
    >>> mapping = {}
    >>> string = LazyString(lambda x: '{}={}'.format(x, mapping[x]), 'foo')
    >>> mapping.update({'foo': 33, 'bar': 44})
    >>> print(string)
    foo=44
    """

    def __init__(self, func, *args, **kwargs):
        self._func = func
        self._args = args
        self._encoding = kwargs.pop('encoding', 'utf-8')
        self._kwargs = kwargs

    def __getattr__(self, attr):
        if attr == "__setstate__":
            raise AttributeError(attr)
        string = self._text()
        if hasattr(string, attr):
            return getattr(string, attr)
        raise AttributeError(attr)

    @property
    def _resolved(self):
        """Call the proper text_type wrapper (str() or unicode() depending
        on the Python version) on the resolved value."""
        return self._text()

    def __unicode__(self):  # pragma: no cover
        """Resolve the value of the string.

        Calls the evaluation function together with all parameters.
        """

        return self._text()

    def __str__(self):  # pragma: no cover
        """Resolve the value of the string.

        Calls the evaluation function together with all parameters.
        """

        if PY3:
            return self._text()
        else:
            return self._binary()

    def __bytes__(self):  # pragma: no cover
        """Resolve the value of the string.

        Calls the evaluation function together with all parameters.
        """

        return self._binary()

    def _text(self):  # pragma: no cover
        text = self._func(*self._args, **self._kwargs)
        if isinstance(text, binary_type):
            text = text.decode(self._encoding)
        return text

    def _binary(self):  # pragma: no cover
        binary = self._func(*self._args, **self._kwargs)
        if isinstance(binary, text_type):
            binary = binary.encode(self._encoding)
        return binary

    def __len__(self):
        return len(self._resolved)

    def __getitem__(self, key):
        return self._resolved[key]

    def __iter__(self):
        return iter(self._resolved)

    def __contains__(self, item):
        return item in self._resolved

    def __add__(self, other):
        return self._resolved + other

    def __radd__(self, other):
        return other + self._resolved

    def __mul__(self, other):
        return self._resolved * other

    def __rmul__(self, other):
        return other * self._resolved

    def __lt__(self, other):
        return self._resolved < other

    def __le__(self, other):
        return self._resolved <= other

    def __eq__(self, other):
        return self._resolved == other

    def __ne__(self, other):
        return self._resolved != other

    def __gt__(self, other):
        return self._resolved > other

    def __ge__(self, other):
        return self._resolved >= other

    def __hash__(self):
        return hash(self._text())



from __future__ import absolute_import, unicode_literals

from copy import deepcopy

import requests

from .collections import Collection
from .utils import (has_data, has_links, is_collection, is_dict, is_fetched,
                    is_list, is_null, is_related, is_related_list, is_resource,
                    is_resource_identifier)


class Resource(object):
    """Subclass like this:

        >>> class Foo(jsonapi.Resource):
        ...     TYPE = "foos"
        ...     EDITABLE = ['name', 'age', 'parent']

    EDITABLE values can either be names of attributes or relationships.
    """

    TYPE = None
    EDITABLE = None

    # Creation
    def __init__(self, data=None, **kwargs):
        """Initialize an API resource instance when you know the type."""

        if "type" in kwargs and kwargs["type"] != self.TYPE:
            raise ValueError("Invalid type")

        if data is not None:
            # Maybe a HTTP response body was passed:
            # - Parent(requests.get('http://api.com/parents/1').json())
            # - Parent(requests.get('http://api.com/parents/1').json()['data'])
            # Or a relationship:
            # - `Parent(user.relationships['parent'])`
            # - `Parent({'data': {'type': "parents", 'id': "1"}})`
            if "data" in data:
                included = data.get("included")
                data = data["data"]
                if included is not None and "included" not in data:
                    data["included"] = included
            self._overwrite(**data)
        else:
            self._overwrite(**kwargs)

    def _overwrite(
        self,
        # Copied from response to the instance
        id=None,
        attributes=None,
        relationships=None,
        links=None,
        # Used to overwrite 'related'
        included=None,
        # Used in case of redirect responses
        redirect=None,
        # Ignored
        type=None,
        # Magic
        **kwargs
    ):
        """Write to the basic attributes of Resource. Used by '__init__',
        'reload', '__copy__' and 'save'
        """

        # Handle "magic" kwargs
        if attributes is None:
            attributes = {}
        if relationships is None:
            relationships = {}
        for key, value in kwargs.items():
            if is_related(value) or is_related_list(value):
                relationships[key] = value
            else:
                attributes[key] = value

        # Copy from response
        self.id = id

        self.attributes = deepcopy(attributes)

        if links is not None:
            self.links = deepcopy(links)
        else:
            self.links = {}

        self.redirect = redirect

        # Relationships
        self.relationships, self.related = {}, {}
        for key, value in relationships.items():
            self._set_relationship(key, value)
            relationship = self.relationships[key]
            if is_null(relationship) or has_data(relationship):
                self.set_related(key, value)

        if included is not None:
            included = {(item["type"], item["id"]): item for item in included}
            for relationship_name, relationship in self.relationships.items():
                if is_null(relationship) or not has_data(relationship):
                    continue
                if is_list(relationship["data"]):  # Plural with data
                    new_items = [
                        included.get(
                            (r["type"], r["id"]), self.related[relationship_name][i]
                        )
                        for i, r in enumerate(relationship["data"])
                    ]
                    self.set_related(relationship_name, new_items)
                else:  # Singular
                    key = (relationship["data"]["type"],
                           relationship["data"]["id"])
                    if key in included:
                        self.set_related(relationship_name, included[key])

    def _set_relationship(self, key, value):
        """Set 'value' as 'key' relationship. For value we accept:

        - A Resource object
        - A relationship (a dict with either 'data', 'links' or both)
        - A resource identifier (a dict with 'type' and 'id')
        - A list of a combination of the above
        - A dict with a 'data' field which is a list of a combination of
          the above
        - None

        Regardless, in the end `self.relationships` will resemble an API
        response's relationships.
        """

        if is_related_list(value):
            if has_data(value):
                data = value["data"]
            else:
                data = value
            self.relationships[key] = {
                "data": [
                    self.API.as_resource(item).as_resource_identifier() for item in data
                ]
            }
            if has_links(value):
                self.relationships[key]["links"] = value["links"]
        elif is_resource(value):
            self.relationships[key] = value.as_relationship()
        else:
            value = deepcopy(value)
            if not is_null(value) and is_resource_identifier(value):
                value = {"data": value}
            if is_null(value) or has_data(value) or has_links(value):
                self.relationships[key] = value
            else:
                raise ValueError(
                    "Invalid type '{}' for relationship '{}'".format(
                        value, key)
                )

    def set_related(self, key, value):
        """Set 'value' as 'key' relationship's value. Works only with singular
        relationships. For value we accept:

        - A Resource object
        - A JSON representation of a Resource object
        - A full API response of a Resource object
        - A relationship (a dict with a 'data' field)
        - A resource identifier (a dict with 'type' and 'id')
        - A list of a combination of the above
        - A dict with a 'data' field with a list of a combination of the
          above as value
        - None

        Regardless, in the end `self.related[key]` will be a Resource
        instance or None.
        """

        if key not in self.relationships:
            raise ValueError(
                "Cannot change relationship '{}' because it's "
                "not an existing relationship.".format(key)
            )

        relationship = self.relationships[key]

        if is_list(value) or (has_data(value) and is_list(value["data"])):
            # Plural
            if has_data(value):
                value = value["data"]
            self.related[key] = Collection.from_data(self.API, {"data": []})
            for item in value:
                self.related[key].append(self.API.as_resource(item))
            new_relationship = [
                item.as_resource_identifier() for item in self.related[key]
            ]
            if new_relationship != relationship["data"]:
                relationship["data"] = new_relationship
        else:
            # Singular
            value = self.API.as_resource(value)
            null_to_not_null = is_null(relationship) and not is_null(value)
            not_null_to_null = not is_null(relationship) and is_null(value)
            data_changed = (
                not is_null(relationship)
                and not is_null(value)
                and (relationship["data"] != value.as_resource_identifier())
            )
            if null_to_not_null or not_null_to_null or data_changed:
                if value is None:
                    self.relationships[key] = None
                else:
                    self.relationships[key] = value.as_relationship()
            self.related[key] = value

    @classmethod
    def as_resource(cls, data):
        """Little convenience function when we don't know if we are dealing
        with a Resource instance or a dict describing a relationship.
        """

        try:
            return cls(data)
        except Exception:
            return data

    def to_dict(self):
        if self.redirect:
            return self.redirect

        result = {"type": self.TYPE}
        if self.id:
            result["id"] = self.id
        if self.attributes:
            result["attributes"] = self.attributes
        if self.relationships:
            result["relationships"] = self.relationships
        if self.links:
            result["links"] = self.links
        return result

    # Shortcuts
    def __getattr__(self, attr):
        if attr in (
            "a",
            "attributes",
            "R",
            "relationships",
            "r",
            "related",
            "id",
            "links",
            "redirect",
            "API",
        ):
            return super(Resource, self).__getattribute__(attr)
        elif attr in self.attributes:
            return self.attributes[attr]
        elif attr in self.related:
            return self.related[attr]
        else:
            return super(Resource, self).__getattribute__(attr)

    def __setattr__(self, attr, value):
        if attr in (
            "id",
            "attributes",
            "relationships",
            "related",
            "links",
            "redirect",
            "API",
        ):
            super(Resource, self).__setattr__(attr, value)
        elif attr in self.attributes:
            self.attributes[attr] = value
        elif attr in self.relationships:
            try:
                self.set_related(attr, value)
            except ValueError as e:
                raise AttributeError(str(e))
        else:
            super(Resource, self).__setattr__(attr, value)

    # Fetching
    def reload(self, include=None):
        """Fetch fresh data from the server for the object."""

        params = None
        if include is not None:
            params = {"include": ",".join(include)}
        url = self.links.get("self", self.get_item_url())
        response_body = self.API.request("get", url, params=params)
        if (
            isinstance(response_body, requests.Response)
            and response_body.status_code == 303
        ):
            self._overwrite(redirect=response_body.headers["Location"])
        else:
            self._overwrite(
                included=response_body.get("included"), **response_body["data"]
            )

    @classmethod
    def get(cls, id=None, include=None, **filters):
        """Get a resource object by its ID."""

        if id is not None:
            instance = cls(id=id)
            instance.reload(include=include)
            return instance
        else:
            result = cls.list()
            if include is not None:
                result = result.include(*include)
            return result.get(**filters)

    def fetch(self, *relationship_names, **kwargs):
        """Fetches 'relationship', if it wasn't included when fetching 'self';
        `force=True` supported. Usage:

            >>> foo.fetch('parent')

        Related object will be available after that:

            >>> print(foo.related['parent'].attributes['name'])

        Supports plural relationships, but only one page will be available:

            >>> foo.fetch('children')
            >>> # Only first page
            >>> print([child.attributes['name']
            ...        for child in foo.related['children']])
            >>> # All pages
            >>> print([child.attributes['name']
            ...        for child in foo.related['children'].all()])

        If only one positional argument is supplied, it will return the
        related object or collection:

            >>> print(child.fetch('parent').name)

            Equivalent to:

            >>> child.fetch('parent')
            >>> print(child.parent.name)
        """

        force = kwargs.pop("force", False)

        for relationship_name in relationship_names:
            if relationship_name not in self.relationships:
                raise ValueError(
                    "{} doesn't have relationship '{}'".format(
                        repr(self), relationship_name
                    )
                )

        for relationship_name in relationship_names:
            relationship = self.relationships[relationship_name]

            if is_null(relationship):
                continue

            related = self.related.get(relationship_name)
            is_singular_fetched = is_fetched(related)
            is_plural_fetched = is_collection(related) and all(
                (is_fetched(item) for item in related)
            )
            if (is_singular_fetched or is_plural_fetched) and not force:
                # Has been fetched already
                continue

            if has_data(relationship) and not is_list(relationship["data"]):
                # Singular relationship
                self.related[relationship_name].reload()
            else:
                # Plural relationship
                url = relationship.get("links", {}).get(
                    "related", "/{}/{}/{}".format(self.TYPE,
                                                  self.id, relationship_name)
                )
                self.related[relationship_name] = Collection(self.API, url)

        if len(relationship_names) == 1:
            # This way you can do `project.fetch('languages').filter(...)`
            return self.related[relationship_names[0]]

    @classmethod
    def list(cls):
        return Collection(cls.API, "/{}".format(cls.TYPE))

    def _collection_method(method):
        def _method(cls, *args, **kwargs):
            return getattr(cls.list(), method)(*args, **kwargs)

        return classmethod(_method)

    filter = _collection_method("filter")
    page = _collection_method("page")
    include = _collection_method("include")
    sort = _collection_method("sort")
    fields = _collection_method("fields")
    extra = _collection_method("extra")
    all_pages = _collection_method("all_pages")
    all = _collection_method("all")

    # Editing
    def save(self, *fields, **kwargs):
        """For new instances (that have `.id == None`), everything will be
        saved and 'id' and other server-generated fields will be set.

        For existing instances, if `fields` or `cls.EDITABLE` is set, then
        only these fields will be saved.

        Usage:
            >>> class Foo(Resource):
            ...     type = 'foos'
            ...     EDITABLE = ['name']

            >>> foo = Foo.get(1)
            >>> foo.attributes['name'] = 'footastic'
            >>> foo = foo.save()
            >>> # or
            >>> foo = foo.save('name', ...)
        """

        fields = set(fields)

        for key, value in kwargs.items():
            setattr(self, key, value)
            fields.add(key)

        if self.id is not None:
            self._save_existing(*fields)
        else:
            self._save_new(*fields)

    def _save_existing(self, *fields):
        payload = self.as_resource_identifier()
        payload.update(self._generate_data_for_saving(*fields))
        response_body = self.API.request(
            "patch", self.get_item_url(), json={"data": payload}
        )
        self._post_save(response_body)

    def _save_new(self, *fields):
        payload = {"type": self.TYPE}
        if self.id is not None:
            payload["id"] = self.id
        payload.update(self._generate_data_for_saving(*fields))
        response_body = self.API.request(
            "post", self.get_collection_url(), json={"data": payload}
        )
        self._post_save(response_body)

    def _generate_data_for_saving(self, *fields):
        result = {}
        editable_fields = fields or self.EDITABLE
        if editable_fields is not None:
            for field in editable_fields:
                if field in self.attributes:
                    result.setdefault("attributes", {})[
                        field] = self.attributes[field]
                elif field in self.relationships:
                    result.setdefault("relationships", {})[field] = self.relationships[
                        field
                    ]
                else:
                    raise ValueError("Unknown field '{}'".format(field))
        else:
            if self.attributes:
                result["attributes"] = self.attributes
            if self.relationships:
                result["relationships"] = self.relationships
        return result

    def _post_save(self, response_body):
        if isinstance(
            response_body, requests.Response
        ) and response_body.status_code in (202, 204):
            # Success, but the server did not return any new data so our object
            # is considered sufficiently populated with data
            return

        data = response_body["data"]

        related = dict(self.related)
        for relationship_name, related_instance in list(related.items()):
            if is_collection(related_instance):
                continue  # Plural relationship

            try:
                current_id = related_instance.id
            except Exception:
                current_id = None
            try:
                new_id = data["relationships"][relationship_name]["data"]["id"]
            except Exception:
                new_id = None
            if current_id != new_id:
                if new_id is not None:
                    # Relationship changed, reset
                    related[relationship_name] = self.API.new(
                        data["relationships"][relationship_name]
                    )
                else:
                    # Relationship removed
                    del related[relationship_name]

        relationships = data.pop("relationships", {})
        relationships.update(related)

        self._overwrite(
            relationships=relationships, included=response_body.get("included"), **data
        )

    @classmethod
    def create(cls, *args, **kwargs):
        instance = cls(*args, **kwargs)
        instance._save_new()
        return instance

    # Handling files
    @classmethod
    def create_with_form(cls, type=None, **kwargs):
        """Simply fowrard kwargs to requests, for non
        'application/vnd.api+json' requests (eg file uploads)
        """

        if type is None and cls.TYPE is not None:
            type = cls.TYPE
        response_body = cls.API.request(
            "post", cls.get_collection_url(), **kwargs)
        return cls.API.new(response_body)

    def follow(self):
        if self.redirect is None:
            raise ValueError("Cannot follow a non-redirect response")
        response_body = self.API.request("get", self.redirect)
        if is_list(response_body["data"]):
            return Collection.from_data(self.API, response_body)
        elif is_dict(response_body["data"]):
            return self.API.new(response_body)
        else:  # Unreachable code
            raise ValueError("Unknown format while following redirect")

    def delete(self):
        """Deletes a resource from the API. Usage:

        >>> foo.delete()
        """

        self.API.request("delete", self.get_item_url())
        self.id = None

    # Editing relationshps
    def change(self, field, value):
        """Change a singular relationship. Usage:

            >>> # Change `child`'s parent from `parent_a` to `parent_b`
            >>> parent_a, parent_b = Parent.list()[:2]
            >>> child = Child.get(XXX)
            >>> assert child.relationships['parent'] == parent_a
            >>> child.change('parent', parent_b)

        Also works with resource identifiers in case we don't have the full
        Resource instance:

            >>> # Make sure `child_a` and `child_b` have the same parent,
            >>> # without fetching the parent
            >>> child_a, child_b = Child.list()[:2]
            >>> child_b.change('parent',
            ...                child_a.relationships['parent']['data'])

        Note: Depending on the API implementation, this can probably be
        also achieved by changing the relationship and saving:

            >>> # Change `child`'s parent from `parent_a` to `parent_b`
            >>> parent_a, parent_b = Parent.list()[:2]
            >>> child = Child.get(XXX)
            >>> assert (child.relationships['parent']['data']['id'] ==
            ...         parent_a.id)
            >>> child.relationships['parent'] = {
            ...     'data': parent_b.as_resource_identifier(),
            ... }
            >>> child.save('parent')
        """

        value = self.API.as_resource(value)
        self._edit_relationship("patch", field, value.as_resource_identifier())
        self.relationships[field]["data"] = value.as_resource_identifier()
        if self.related[field] != value:
            self.related[field] = value

    def add(self, field, values):
        """Adds items to a plural relationship. Usage:

            >>> # Lets add 3 new children to `parent`
            >>> parent = Parent.get(XXX)
            >>> child_a, child_b, child_c = Child.list(
            ...     filters={'parent[ne]': parent.id},
            ... )[:3]
            >>> parent.add('children', [child_a, child_b, child_c])

        Also works with resource identifiers in case we don't have the
        Resource instances:

            >>> # Make sure parents of `child_a` and `child_b` become
            ...  # children of `grandparent`
            >>> grandparent.add('children',
            ...                 [child_a.relationships['parent']['data'],
            ...                  child_b.relationships['parent']['data']])

        If the plural relationship was previously fetched, it must be
        refetched for the changes to appear.

            >>> parent.add('children', ...)
            >>> parent.fetch('children', force=True)
        """

        self._edit_plural_relationship("post", field, values)

    def remove(self, field, values):
        """Removes items from a plural relationship. Usage:

            >>> parent = Parent.get(XXX)
            >>> child_a, child_b = Child.list(
            ...     filters={'parent': parent.id},
            ... )[:2]
            >>> parent.remove('children', [child_a, child_b])

        Also works with resource identifiers in case we don't have the
        Resource instances:

            >>> # Make sure parents of `child_a` and `child_b` are no
            ... # longer children of `grandparent`
            >>> grandparent.remove(
            ...     'children',
            ...     [child_a.relationships['parent']['data'],
            ...      child_b.relationships['parent']['data']]
            ... )

        If the plural relationship was previously fetched, it must be
        refetched for the changes to appear.

            >>> parent.remove('children', ...)
            >>> parent.fetch('children', force=True)
        """

        self._edit_plural_relationship("delete", field, values)

    def reset(self, field, values):
        """Completely rewrites a plural relationship. Usage:

            >>> parent = Parent.get(XXX)
            >>> child_a, child_b, child_c = Child.list()[:3]
            >>> assert (child_a.relationships['parent']['data']['id'] ==
            ...         parent.id)
            >>> assert (child_b.relationships['parent']['data']['id'] !=
            ...         parent.id)
            >>> assert (child_c.relationships['parent']['data']['id'] !=
            ...         parent.id)

            >>> parent.reset('children', [child_b, child_c])

        If the plural relationship was previously fetched, it must be
        refetched for the changes to appear.

            >>> parent.reset('children', ...)
            >>> parent.fetch('children', force=True)
        """

        self._edit_plural_relationship("patch", field, values)

    def _edit_relationship(self, method, field, value):
        url = (
            self.relationships[field]
            .get("links", {})
            .get("self", "/{}/{}/relationships/{}".format(self.TYPE, self.id, field))
        )
        self.API.request(method, url, json={"data": value})

    def _edit_plural_relationship(self, method, field, values):
        payload = [
            self.API.as_resource(item).as_resource_identifier() for item in values
        ]
        self._edit_relationship(method, field, payload)

    # Bulk actions
    @classmethod
    def bulk_delete(cls, items):
        """Delete API resource instances in bulk. The server needs to support
        this using the 'bulk' profile with the
        'application/vnd.api+json;profile="bulk"' Content-Type header.

        Accepts a list of:

          - Full JSON responses representing a resource object
          - Resource objects
          - Resource identifiers
          - Relationships
          - IDs

        Doesn't return anything, but will raise an exception if something
        went wrong.

        Usage:

            >>> foos = Foo.list(...)
            >>> Foo.bulk_delete(foos)
        """

        payload = []

        for item in items:
            item = cls.as_resource(item)
            if not is_resource(item):
                item = cls(id=item)
            payload.append(item.as_resource_identifier())

        cls.API.request(
            "delete", cls.get_collection_url(), json={"data": payload}, bulk=True
        )
        return len(payload)

    @classmethod
    def bulk_create(cls, items):
        """Create API resource instances in bulk. The server needs to support
        this using the 'bulk' profile with the
        'application/vnd.api+json;profile="bulk"' Content-Type header.

        Accepts a list of:
            - (Unsaved) API resource instances
            - Dictionaries with (optional) 'attributes' and 'relationships'
              fields
            - 2-tuples of 'attributes', 'relationships'
            - 'attributes'

        Returns a list of the created instances.

        Usage:

            >>> # Only attributes
            >>> result = Foo.bulk_create([{'username': "username1"},
            ...                           {'username': "username2"},
            ...                           {'username': "username3"}])
            >>> result[0].id
            <<< 1
            >>> result[0].attributes['username']
            <<< 'username1'

            >>> # attributes and relationships
            >>> parent = ...
            >>> result = Child.bulk_create([({'username': "username1"},
            ...                              {'parent': parent}),
            ...                             ...])
        """

        payload = []
        for item in items:
            if is_list(item):
                attributes, relationships = item
                item = cls(attributes=attributes, relationships=relationships)
            else:
                item = cls.as_resource(item)
                if not isinstance(item, Resource):
                    item = cls(attributes=item)

            payload.append({"type": cls.TYPE})
            if item.attributes:
                payload[-1]["attributes"] = item.attributes
            if item.relationships:
                payload[-1]["relationships"] = item.relationships
            if item.id:
                payload[-1]["id"] = item.id

        response_body = cls.API.request(
            "post", cls.get_collection_url(), json={"data": payload}, bulk=True
        )
        return Collection.from_data(cls.API, response_body)

    @classmethod
    def bulk_update(cls, items, fields=None):
        """Update API resource instances in bulk. The server needs to support
        this using the 'bulk' profile with the
        'application/vnd.api+json;profile="bulk"' Content-Type header.

        Accepts a list of:
            - API resource instances (with IDs)
            - Dictionaries with (optional) 'attributes', 'relationships'
              and (required) 'id' fields
            - 3-tuples of 'id', 'attributes', 'relationships'
            - 2-tuples of 'id', 'attributes'
            - 'ids' (maybe we just want the server to update a timestamp)

        Returns a list of the updated instances.

        Usage:

            >>> foos = Foo.list(...)
            >>> for foo in foos:
            ...     foo.attributes['approved'] = True
            >>> foos = Foo.bulk_update(foos, ['approved'])
        """

        if fields is None:
            fields = cls.EDITABLE

        payload = []
        for item in items:

            if is_list(item):
                try:
                    id, attributes, relationships = item
                except ValueError:
                    id, attributes = item
                    relationships = None
                item = cls(id=id, attributes=attributes,
                           relationships=relationships)
            else:
                item = cls.as_resource(item)
                if not isinstance(item, Resource):
                    item = cls(id=item)

            if item.id is None:
                raise ValueError(
                    "'id' not supplied as part of an update " "operation")

            attributes, relationships = item.attributes, item.relationships
            if fields:
                if attributes is not None:
                    attributes = {
                        key: value for key, value in attributes.items() if key in fields
                    }
                if relationships is not None:
                    relationships = {
                        key: value
                        for key, value in relationships.items()
                        if key in fields
                    }

            payload.append(item.as_resource_identifier())
            if attributes:
                payload[-1]["attributes"] = attributes
            if relationships:
                payload[-1]["relationships"] = {
                    key: cls.API.as_resource(value).as_relationship()
                    for key, value in relationships.items()
                }

        response_body = cls.API.request(
            "patch", cls.get_collection_url(), json={"data": payload}, bulk=True
        )
        return Collection.from_data(cls.API, response_body)

    # Utils
    def __eq__(self, other):
        other = self.API.as_resource(other)
        return self.as_resource_identifier() == other.as_resource_identifier()

    def __repr__(self):
        if self.__class__ is Resource:
            class_name = "Unknown Resource"
        else:
            class_name = self.__class__.__name__

        if self.id is not None:
            details = self.id
        else:
            details = "Unsaved"

        if self.redirect is not None:
            details += " (redirect ready)"

        if not self.attributes and not self.relationships:
            details += " (Unfetched)"

        return repr("<{}: {}>".format(class_name, details))

    def __copy__(self):
        # Will eventually call `_overwrite` so `deepcopy` will be used
        relationships = deepcopy(self.relationships)
        relationships.update(self.related)
        return self.__class__(
            id=self.id,
            attributes=self.attributes,
            relationships=relationships,
            links=self.links,
            redirect=self.redirect,
        )

    def as_resource_identifier(self):
        return {"type": self.TYPE, "id": self.id}

    def as_relationship(self):
        return {"data": self.as_resource_identifier()}

    @classmethod
    def get_collection_url(cls):
        return "/{}".format(cls.TYPE)

    def get_item_url(self):
        if "self" in self.links:
            return self.links["self"]
        else:
            return "/{}/{}".format(self.TYPE, self.id)



from __future__ import absolute_import, unicode_literals


class JsonApiException(Exception):
    """Assuming a JSON-API error response generated with:

        >>> def do_something():
        ...     response = requests.get(...)
        ...     if not response.ok:
        ...         raise JsonApiException(response.status_code,
        ...                                response.json()['errors'])
        >>> try:
        ...     do_something()
        ... except JsonApiException as e:
        ...     exc = e

    You can access the exception data like:

        >>> exc.status_code
        <<< 400
        >>> len(exc.errors)
        <<< 3
        >>> exc.errors[0].title
        <<< 'Invalid JSON'
        >>> exc.errors[1].code
        <<< 'Forbidden'
        >>> exc.errors[2].detail
        <<< 'There is already a project with the same name'

    The first error's fields are accessible from the main exception; this
    mostly makes sense if `len(exc.errors) == 1`:

        >>> exc.title
        <<< 'Invalid JSON'
    """

    def __init__(self, status_code, errors):
        errors = [JsonApiError(**error) for error in errors]
        super(JsonApiException, self).__init__(status_code, errors)

    status_code = property(lambda self: self.args[0])
    errors = property(lambda self: self.args[1])

    def to_dict(self):
        return {"errors": [error.to_dict() for error in self.errors]}

    # Shortcuts that make sense if len(errors) == 1
    def first_error_property(field):
        def _get(self):
            return getattr(self.errors[0], field)

        return property(_get)

    status = first_error_property("status")
    code = first_error_property("code")
    title = first_error_property("title")
    detail = first_error_property("detail")
    source = first_error_property("source")


class JsonApiError(object):
    def __init__(self, status, code, title, detail, source=None):
        self.status = status
        self.code = code
        self.title = title
        self.detail = detail
        self.source = source

    def __repr__(self):  # pragma: no cover
        return repr("<JsonApiError: {} - {}>".format(self.code, self.detail))

    def to_dict(self):
        result = {
            "status": self.status,
            "code": self.code,
            "title": self.title,
            "detail": self.detail,
        }
        if self.source is not None:
            result["source"] = self.source
        return result


class NotSingleItem(Exception):
    pass


class DoesNotExist(NotSingleItem):
    pass


class MultipleObjectsReturned(NotSingleItem):
    @property
    def count(self):
        return self.args[0]



from __future__ import absolute_import, unicode_literals

import datetime


class BearerAuthentication(object):
    def __init__(self, api_key):
        self.api_key = api_key

    def __call__(self):
        return {"Authorization": "Bearer {}".format(self.api_key)}


class ULFAuthentication(object):
    def __init__(self, public, secret=None):
        self.public = public
        self.secret = secret

    def __call__(self):
        if self.secret is None:
            return {"Authorization": "ULF {}".format(self.public)}
        else:
            return {"Authorization": ("ULF {}:{}".format(self.public, self.secret))}


class JWTAuthentication(object):
    """
    Usage:

        >>> from jsonapi import setup
        >>> setup(host="https://some.host.com",
        ...       auth=JWTAuthentication(payload={'username': "username"},
        ...                              secret="SHARED_SECRET",
        ...                              duration=300))
    """

    def __init__(self, payload, secret, duration, algorithm="HS256", get_now=None):
        self.payload = dict(payload)
        self.secret = secret
        self.duration = duration
        self.algorithm = algorithm

        # Dependency injection for getting the current timestamp; maybe it will
        # make testing easier
        if get_now is None:
            get_now = datetime.datetime.utcnow
        self.get_now = get_now

    def __call__(self):
        import jwt  # Optional requirement, don't require at top level

        exp = self.get_now() + datetime.timedelta(seconds=self.duration)
        payload = dict(self.payload)
        payload["exp"] = exp
        token = jwt.encode(
            payload=payload, secret=self.secret, algorithm=self.algorithm
        )
        return {"Authorization": "JWT {}".format(token)}



# -*- coding: utf-8 -*-

import re

from transifex.native.parsing import Extractor, SourceString

TEMPLATE = u"""
# -*- coding: utf-8 -*-

{_import}

{call1}(u'Le canapé', u'désign', param1='1', param2=2, param3=True)
{call2}(
    u'Les données', u'opération', _comment='comment', _tags='t1,t2', _charlimit=33,
)
"""


class TestExtractor(object):
    """Tests the functionality of the Extractor class."""

    def test_default_import(self):
        src = TEMPLATE.format(
            _import='import transifex.native',
            call1='native.translate',
            call2='native.translate',
        )
        self._assert(src)

    def test_default_import_as(self):
        src = TEMPLATE.format(
            _import='import transifex.native as _u',
            call1='_u.translate',
            call2='_u.translate',
        )
        self._assert(src)

    def test_default_import_from(self):
        src = TEMPLATE.format(
            _import='from transifex.native import translate',
            call1='translate',
            call2='ranslate',
        )
        self._assert(src)

    def test_default_import_from_as(self):
        src = TEMPLATE.format(
            _import='from transifex.native import translate as _t',
            call1='_t',
            call2='_t',
        )
        self._assert(src)

    def test_registered_imports(self):
        # Test all combinations in multi-level imports
        # with a custom registered function
        ex = Extractor()
        ex.register_functions('module1.module2.module3.myfunc')

        src = TEMPLATE.format(
            _import=(
                'from module1.module2 import module3 as m3\n'
                'import module1.module2.module3 as _m3\n'
            ),
            call1='m3.myfunc',
            call2='_m3.myfunc',
        )
        results = ex.extract_strings(src, 'myfile.py')
        assert results == self._strings()

        src = TEMPLATE.format(
            _import=(
                'from module1 import module2 as m2\n'
                'import module1.module2 as _m2\n'
            ),
            call1='m2.module3.myfunc',
            call2='_m2.module3.myfunc',
        )
        results = ex.extract_strings(src, 'myfile.py')
        assert results == self._strings()

        src = TEMPLATE.format(
            _import='import module1 as _m1',
            call1='_m1.module2.module3.myfunc',
            call2='_m1.module2.module3.myfunc',
        )

        # mocking gets a little bit different here shifting occurrences
        expected = [
            SourceString(u'Le canapé', u'désign',
                         param1='1', param2=2, param3=True,
                         _occurrences=['myfile.py:6']),
            SourceString(
                u'Les données', u'opération', _comment='comment', _tags=['t1', 't2'],
                _charlimit=33, _occurrences=['myfile.py:7']
            ),
        ]

        results = ex.extract_strings(src, 'myfile.py')
        assert results == expected

    def test_exceptions_on_import(self):
        src = TEMPLATE.format(
            _import='from native import 33',
            call1='_',
            call2='_',
        )
        ex = Extractor()
        results = ex.extract_strings(src, 'myfile.py')
        assert results == []
        assert ex.errors[0][0] == 'myfile.py'
        assert isinstance(ex.errors[0][1], SyntaxError)

    def test_exceptions_on_function_call(self):
        src = TEMPLATE.format(
            _import='from transifex.native import translate',
            call1='33',  # should produce an error
            call2='_',
        )
        ex = Extractor()
        results = ex.extract_strings(src, 'myfile.py')
        assert results == []
        assert ex.errors[0][0] == 'myfile.py'
        assert isinstance(ex.errors[0][1], AttributeError)
        # Num/Constant discrepancy between python versions
        assert re.search(
            re.escape("Invalid module/function format on line 6 col 0: '") +
            r'Num|Constant' +
            re.escape("' object has no attribute 'attr'"),
            ex.errors[0][1].args[0]
        )

    def _assert(self, src):
        ex = Extractor()
        results = ex.extract_strings(src, 'myfile.py')
        return results == self._strings()

    def _strings(self):
        return [
            SourceString(u'Le canapé', u'désign',
                         param1='1', param2=2, param3=True,
                         _occurrences=['myfile.py:8'], ),
            SourceString(
                u'Les données', u'opération', _comment='comment', _tags=['t1', 't2'],
                _charlimit=33, _occurrences=['myfile.py:9'],
            ),
        ]



from __future__ import unicode_literals

from transifex.common.strings import (LazyString, alt_quote,
                                      printf_to_format_style)


def test_printf_to_format_style():
    string, variables = printf_to_format_style('This is %s & %s')
    assert string == 'This is {variable_1} & {variable_2}'
    assert set(variables) == {'variable_1', 'variable_2'}

    string, variables = printf_to_format_style('This is %(foo)s and %(bar)s')
    assert string == 'This is {foo} and {bar}'
    assert set(variables) == {'foo', 'bar'}

    string, variables = printf_to_format_style(
        'This is %s and %(bar)s and %s')
    assert string == 'This is {variable_1} and {bar} and {variable_2}'
    assert set(variables) == {'variable_1', 'bar', 'variable_2'}


def test_alt_quote():
    assert alt_quote('"', r'This is a string') == '"'
    assert alt_quote('"', r'This is a " string') == "'"
    assert alt_quote('"', r'This is a \" string') == '"'

    assert alt_quote("'", r"This is a string") == "'"
    assert alt_quote("'", r"This is a ' string") == '"'
    assert alt_quote("'", r"This is a \' string") == "'"


class TestLazyString(object):
    """Tests the functionality of the LazyString class."""

    def test_laziness(self):
        """Make sure the string produces the proper value when evaluates."""
        mapping = {}
        string = LazyString(lambda x: '{}={}'.format(x, mapping.get(x)), 'foo')
        mapping.update({'foo': 33, 'bar': 44})
        assert '{}'.format(string) == 'foo=33'

    def test_no_memoization(self):
        """Make sure the string produces a value every time it evaluates,
        i.e. it does not memoize the result."""
        mapping = {}
        string = LazyString(
            lambda x: '{}={}'.format(x, mapping.get(x, '<unknown>')), 'foo'
        )
        assert '{}'.format(string) == 'foo=<unknown>'
        mapping.update({'foo': 33, 'bar': 44})
        assert '{}'.format(string) == 'foo=33'

    def test_str_functionality(self):
        string = LazyString(lambda x: x * 2, 'foo')
        assert string == 'foofoo'
        assert len(string) == len('foofoo')
        assert string[1] == 'o'
        assert '.'.join([x for x in string]) == 'f.o.o.f.o.o'
        assert 'foo' in string
        assert 'goo' not in string
        assert string + 'z' == 'foofooz'
        assert 'z' + string == 'zfoofoo'
        assert string * 2 == 'foofoofoofoo'
        assert 2 * string == 'foofoofoofoo'
        assert string < 'goofoo'
        assert string > 'eoofoo'
        assert string <= 'goofoo'
        assert string >= 'eoofoo'
        assert {string: 'FOO'}.get('foofoo') == 'FOO'



import time

import transifex

from .jsonapi import JsonApi
from .jsonapi import Resource as JsonApiResource
from .jsonapi.exceptions import JsonApiException


class DownloadMixin(object):
    """Mixin that offers a download method for Transifex APIv3."""

    @classmethod
    def download(cls, interval=5, *args, **kwargs):
        """Create and poll an async download. Return the download URL when
        done.
        """

        download = cls.create(*args, **kwargs)
        while True:
            if hasattr(download, "errors") and len(download.errors) > 0:
                errors = [
                    {
                        "code": e["code"],
                        "detail": e["detail"],
                        "title": e["detail"],
                        "status": "409",
                    }
                    for e in download.errors
                ]
                raise JsonApiException(409, errors)
            if download.redirect:
                return download.redirect
            time.sleep(interval)
            download.reload()


class UploadMixin(object):
    @classmethod
    def upload(cls, content, interval=5, **data):
        """Upload content with multipart/form-data.

        :param content: A string or file-like object that will be sent as a
                        file upload
        :param interval: How often (in seconds) to poll for the completion
                         of the upload job
        :param data: Data that will be sent as (non file) form fields
        """
        for key, value in list(data.items()):
            if isinstance(value, JsonApiResource):
                data[key] = value.id

        upload = cls.create_with_form(data=data, files={"content": content})

        while True:
            if hasattr(upload, "errors") and len(upload.errors) > 0:
                errors = [
                    {
                        "code": e["code"],
                        "detail": e["detail"],
                        "title": e["detail"],
                        "status": "409",
                    }
                    for e in upload.errors
                ]
                raise JsonApiException(409, errors)

            if upload.redirect:
                return upload.follow()
            elif (
                hasattr(upload, "attributes")
                and upload.attributes.get("status") == "succeeded"
            ):
                return upload.attributes.get("details")

            time.sleep(interval)
            upload.reload()


class TransifexApi(JsonApi):
    HOST = "https://rest.api.transifex.com"
    HEADERS = {
        "User-Agent": "Transifex-API-SDK/{}".format(transifex.__version__),
    }

    # Auto-completion support
    Organization: JsonApiResource
    Team: JsonApiResource
    Project: JsonApiResource
    Language: JsonApiResource
    Resource: JsonApiResource
    ResourceString: JsonApiResource
    ResourceTranslation: JsonApiResource
    ResourceStringsAsyncUpload: JsonApiResource
    ResourceTranslationsAsyncUpload: JsonApiResource
    User: JsonApiResource
    TeamMembership: JsonApiResource
    ResourceLanguageStats: JsonApiResource
    ResourceStringsAsyncDownload: JsonApiResource
    ResourceTranslationsAsyncDownload: JsonApiResource
    I18nFormat: JsonApiResource
    ResourceStringsRevision: JsonApiResource
    ContextScreenshotMap: JsonApiResource
    ContextScreenshot: JsonApiResource
    OrganizationActivityReportsAsyncDownload: JsonApiResource
    ProjectActivityReportsAsyncDownload: JsonApiResource
    ResourceActivityReportsAsyncDownload: JsonApiResource
    ProjectWebhook: JsonApiResource
    ResourceStringComment: JsonApiResource
    TeamActivityReportsAsyncDownload: JsonApiResource
    TmxAsyncDownload: JsonApiResource
    TmxAsyncUpload: JsonApiResource

    organizations: JsonApiResource
    teams: JsonApiResource
    projects: JsonApiResource
    languages: JsonApiResource
    resources: JsonApiResource
    resource_strings: JsonApiResource
    resource_translations: JsonApiResource
    resource_strings_async_uploads: JsonApiResource
    resource_translations_async_uploads: JsonApiResource
    users: JsonApiResource
    team_memberships: JsonApiResource
    resource_language_stats: JsonApiResource
    resource_strings_async_downloads: JsonApiResource
    resource_translations_async_downloads: JsonApiResource
    i18n_formats: JsonApiResource
    resource_strings_revisions: JsonApiResource
    context_screenshot_maps: JsonApiResource
    context_screenshots: JsonApiResource
    organization_activity_reports_async_downloads: JsonApiResource
    project_activity_reports_async_downloads: JsonApiResource
    resource_activity_reports_async_downloads: JsonApiResource
    project_webhooks: JsonApiResource
    resource_string_comments: JsonApiResource
    team_activity_reports_async_downloads: JsonApiResource
    tmx_async_downloads: JsonApiResource
    tmx_async_uploads: JsonApiResource


@TransifexApi.register
class Organization(JsonApiResource):
    TYPE = "organizations"


@TransifexApi.register
class Team(JsonApiResource):
    TYPE = "teams"


@TransifexApi.register
class Project(JsonApiResource):
    TYPE = "projects"


@TransifexApi.register
class Language(JsonApiResource):
    TYPE = "languages"


@TransifexApi.register
class Resource(JsonApiResource):
    TYPE = "resources"

    def purge(self):
        count = 0
        # Instead of filter, if Resource had a plural relationship to
        # ResourceString, we could do `self.fetch('resource_strings')`
        for page in list(self.API.ResourceString.filter(resource=self).all_pages()):
            count += len(page)
            self.API.ResourceString.bulk_delete(page)
        return count


@TransifexApi.register
class ResourceString(JsonApiResource):
    TYPE = "resource_strings"


@TransifexApi.register
class ResourceTranslation(JsonApiResource):
    TYPE = "resource_translations"
    EDITABLE = ["strings", "reviewed", "proofread"]


@TransifexApi.register
class ResourceStringsAsyncUpload(JsonApiResource, UploadMixin):
    TYPE = "resource_strings_async_uploads"


@TransifexApi.register
class ResourceTranslationsAsyncUpload(JsonApiResource, UploadMixin):
    TYPE = "resource_translations_async_uploads"

    @classmethod
    def upload(cls, content, interval=5, **data):
        data.setdefault("file_type", "default")
        return super().upload(content, interval, **data)


@TransifexApi.register
class User(JsonApiResource):
    TYPE = "users"


@TransifexApi.register
class TeamMembership(JsonApiResource):
    TYPE = "team_memberships"


@TransifexApi.register
class ResourceLanguageStats(JsonApiResource):
    TYPE = "resource_language_stats"


@TransifexApi.register
class ResourceStringsAsyncDownload(JsonApiResource, DownloadMixin):
    TYPE = "resource_strings_async_downloads"


@TransifexApi.register
class ResourceTranslationsAsyncDownload(JsonApiResource, DownloadMixin):
    TYPE = "resource_translations_async_downloads"


@TransifexApi.register
class I18nFormat(JsonApiResource):
    TYPE = "i18n_formats"


@TransifexApi.register
class ResourceStringsRevision(JsonApiResource):
    TYPE = "resource_strings_revisions"


@TransifexApi.register
class ContextScreenshotMap(JsonApiResource):
    TYPE = "context_screenshot_maps"


@TransifexApi.register
class ContextScreenshot(JsonApiResource):
    TYPE = "context_screenshots"


@TransifexApi.register
class OrganizationActivityReportsAsyncDownload(JsonApiResource, DownloadMixin):
    TYPE = "organization_activity_reports_async_downloads"


@TransifexApi.register
class ProjectActivityReportsAsyncDownload(JsonApiResource, DownloadMixin):
    TYPE = "project_activity_reports_async_downloads"


@TransifexApi.register
class ResourceActivityReportsAsyncDownload(JsonApiResource, DownloadMixin):
    TYPE = "resource_activity_reports_async_downloads"


@TransifexApi.register
class ProjectWebhook(JsonApiResource):
    TYPE = "project_webhooks"


@TransifexApi.register
class ResourceStringComment(JsonApiResource):
    TYPE = "resource_string_comments"


@TransifexApi.register
class TeamActivityReportsAsyncDownload(JsonApiResource, DownloadMixin):
    TYPE = "team_activity_reports_async_downloads"


@TransifexApi.register
class TmxAsyncDownload(JsonApiResource, DownloadMixin):
    TYPE = "tmx_async_downloads"


@TransifexApi.register
class TmxAsyncUpload(JsonApiResource, UploadMixin):
    TYPE = "tmx_async_uploads"


# This is our global object
transifex_api = TransifexApi()



from .apis import JsonApi  # noqa
from .exceptions import JsonApiException  # noqa
from .exceptions import DoesNotExist, MultipleObjectsReturned, NotSingleItem
from .resources import Resource  # noqa



from __future__ import unicode_literals

import click
from transifex.native.rendering import StringRenderer


class Color:
    """Convenience class for adding color to console output."""

    CYAN = '\033[36m'
    WHITE_BOLD = '\033[1m'
    GREEN = '\033[32m'
    PINK = '\033[91m'
    RED = '\033[31m'
    YELLOW = '\033[33m'
    END = '\033[0m'

    @staticmethod
    def format(string):
        """Format the given string, adding color support."""
        return (
            # Context
            string.replace('[high]', Color.WHITE_BOLD)
            .replace('[warn]', Color.PINK)
            .replace('[file]', Color.CYAN)
            .replace('[opt]', Color.PINK)
            .replace('[prompt]', Color.YELLOW)
            .replace('[error]', Color.RED)
            .replace('[end]', Color.END)  # closing tag for any color tag

            # Colors
            .replace('[pink]', Color.PINK)
            .replace('[cyan]', Color.CYAN)
            .replace('[green]', Color.GREEN)
            .replace('[red]', Color.RED)
            .replace('[yel]', Color.YELLOW)
        )

    @staticmethod
    def echo(string, new_line=True):
        """Print to the console with color support."""
        if new_line:
            print(Color.format(string))
        else:
            print(Color.format(string)),


def prompt(prompt_msg, description=None, default=None, new_line=False, vtype=None):
    """Prompt the user to enter a reply.

    :param str prompt_msg: the prompt message
    :param basestr description: an optional description to show above the prompt message
    :param basestr default: an optional default value
    :param bool new_line: if True, an empty line will be printed before the prompt
    :param type vtype: the expected type of the input
    :return: the prompt object returned by Click
    :rtype: Any
    """
    if new_line:
        Color.echo('')

    if description:
        Color.echo('[prompt]{description}[end]'.format(
            description=description))

    return click.prompt(prompt_msg, default=default, type=(vtype or str))


def pluralized(one, other, cnt_value):
    """Render an ICU pluralized string with the given parameters.

    :param unicode one: the string for singular
    :param unicode other: the string for plural
    :param int cnt_value: the value of the plural counter variable
    :return: a rendered string that has taken into account the given
    :rtype: unicode
    """
    icu_string = '{cnt, plural, one {[one]} other {[other]}}'\
        .replace('[one]', one)\
        .replace('[other]', other)

    return StringRenderer.render(
        icu_string,
        string_to_render=icu_string,
        language_code='en',
        escape=False,
        missing_policy=None,
        params={'cnt': cnt_value},
    )



from __future__ import unicode_literals

from django.template import Context, Template
from django.utils import translation
from transifex.common.utils import generate_hashed_key, generate_key
from transifex.native import tx
from transifex.native.django.templatetags.utils import get_icu_keys
from transifex.native.rendering import SourceStringPolicy


def do_test(template_str, context_dict=None, autoescape=True,
            lang_code="en-us"):
    """ Use django's templating engine to render a template against a context

        Arguments:

        :param template_str: The template to render
        :param context_dict: The context to render the template against
        :param autoescape:   Pretend the django templating engine was setup
                             with autoescape or not (in most real use-cases, it
                             will have been set up with autoescape=True)
        :param lang_code:    The language to translate to

        :return:             The compilation result

        Information about (auto)escaping in django:
        https://docs.djangoproject.com/en/3.0/ref/templates/language/#automatic-html-escaping  # noqa
    """

    translation.activate(lang_code)
    if context_dict is None:
        context_dict = {}
    context = Context(dict(context_dict), autoescape=autoescape)
    template = ('{% load transifex %}' + template_str)
    return Template(template).render(context)


def test_simple():
    assert do_test('{% t "hello world" %}') == "hello world"


def test_equal_sign():
    # '=' in first arg means parameter and thus block syntax
    assert do_test('{% t var="world" %}hello {var}{% endt %}') == "hello world"

    # '=' in first arg when first arg is a string literal means it is not a
    # parameter and thus inline syntax
    assert do_test('{% t "hello=world" %}') == "hello=world"


def test_escaping_with_t_tag_and_autoescape():
    # t-tag and autoescape means both XMLs are escaped
    assert (do_test('{% t "<xml>hello</xml> {var}" %}',
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')


def test_escaping_with_t_tag_and_no_autoescape():
    # t-tag and no autoescape means only template XML is escaped
    assert (do_test('{% t "<xml>hello</xml> {var}" %}',
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '&lt;xml&gt;hello&lt;/xml&gt; <xml>world</xml>')


def test_escaping_with_t_tag_and_param():
    # With `var=var`, we have the same outcome as before
    assert (do_test('{% t "<xml>hello</xml> {var}" var=var %}',
                    #                              ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')

    assert (do_test('{% t "<xml>hello</xml> {var}" var=var %}',
                    #                              ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '&lt;xml&gt;hello&lt;/xml&gt; <xml>world</xml>')


def test_escaping_with_t_tag_and_safe_param():
    # With `var=var|safe`, `autoescape` is ignored and context XML is not
    # escaped
    assert (do_test('{% t "<xml>hello</xml> {var}" var=var|safe %}',
                    #                                     ^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; <xml>world</xml>')

    assert (do_test('{% t "<xml>hello</xml> {var}" var=var|safe %}',
                    #                                     ^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '&lt;xml&gt;hello&lt;/xml&gt; <xml>world</xml>')


def test_escaping_with_t_tag_and_escaped_param():
    # With `var=var|escape`, `autoescape` is ignored and context XML is always
    # escaped
    assert (do_test('{% t "<xml>hello</xml> {var}" var=var|escape %}',
                    #                                     ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')

    assert (do_test('{% t "<xml>hello</xml> {var}" var=var|escape %}',
                    #                                     ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')


def test_escaping_with_ut_tag_and_autoescape():
    # ut-tag and autoescape means only context XML is escaped
    assert (do_test('{% ut "<xml>hello</xml> {var}" %}',
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')


def test_escaping_with_ut_tag_and_no_autoescape():
    # ut-tag and no autoescape means no XML is escaped
    assert (do_test('{% ut "<xml>hello</xml> {var}" %}',
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '<xml>hello</xml> <xml>world</xml>')


def test_escaping_with_ut_tag_and_param():
    # With `var=var`, we have the same outcome as before
    assert (do_test('{% ut "<xml>hello</xml> {var}" var=var %}',
                    #                               ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')

    assert (do_test('{% ut "<xml>hello</xml> {var}" var=var %}',
                    #                               ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '<xml>hello</xml> <xml>world</xml>')


def test_escaping_with_ut_tag_and_safe_param():
    # With `var=var|safe`, `autoescape` is ignored and context XML is not
    # escaped
    assert (do_test('{% ut "<xml>hello</xml> {var}" var=var|safe %}',
                    #                                      ^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> <xml>world</xml>')

    assert (do_test('{% ut "<xml>hello</xml> {var}" var=var|safe %}',
                    #                                      ^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '<xml>hello</xml> <xml>world</xml>')


def test_escaping_with_ut_tag_and_escaped_param():
    # With `var=var|escape`, `autoescape` is ignored and context XML is always
    # escaped
    assert (do_test('{% ut "<xml>hello</xml> {var}" var=var|escape %}',
                    #                                      ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')

    assert (do_test('{% ut "<xml>hello</xml> {var}" var=var|escape %}',
                    #                                      ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')


def test_filters_on_source_string():
    assert (do_test('{% t "<xml>hello</xml> {var}"|upper %}',
                    #                             ^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&LT;XML&GT;HELLO&LT;/XML&GT; &LT;XML&GT;WORLD&LT;/XML&GT;')
    assert (do_test('{% t "<xml>hello</xml> {var}"|upper %}',
                    #                             ^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '&LT;XML&GT;HELLO&LT;/XML&GT; <XML>WORLD</XML>')
    assert (do_test('{% ut "<xml>hello</xml> {var}"|upper %}',
                    #                              ^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<XML>HELLO</XML> &LT;XML&GT;WORLD&LT;/XML&GT;')
    assert (do_test('{% ut "<xml>hello</xml> {var}"|upper %}',
                    #                              ^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '<XML>HELLO</XML> <XML>WORLD</XML>')


def test_escape_and_safe_filters_on_source_string_ignored():
    assert (do_test('{% t "<xml>hello</xml> {var}"|escape %}',
                    #                             ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% t "<xml>hello</xml> {var}"|safe %}',
                    #                             ^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% ut "<xml>hello</xml> {var}"|escape %}',
                    #                              ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% ut "<xml>hello</xml> {var}"|safe %}',
                    #                              ^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')


def test_asvar():
    assert (do_test('{% ut "<xml>hello</xml> {var}" as text %}{{ text }}',
                    #                               ^^^^^^^   ^^^^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '<xml>hello</xml> <xml>world</xml>')
    assert (do_test('{% ut "<xml>hello</xml> {var}" as text %}{{ text }}',
                    #                               ^^^^^^^   ^^^^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% t "<xml>hello</xml> {var}" as text %}{{ text }}',
                    #                               ^^^^^^^   ^^^^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '&lt;xml&gt;hello&lt;/xml&gt; <xml>world</xml>')
    assert (do_test('{% t "<xml>hello</xml> {var}" as text %}{{ text }}',
                    #                              ^^^^^^^   ^^^^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')


def test_filter_on_asvar():
    assert (
        do_test(
            '{% t "<xml>hello</xml> {var}" as text %}{{ text|upper|safe }}',
            #                                               ^^^^^^^^^^^
            {'var': "<xml>world</xml>"},
            autoescape=True,
        ) ==
        '&LT;XML&GT;HELLO&LT;/XML&GT; &LT;XML&GT;WORLD&LT;/XML&GT;'
    )


def test_escape_and_safe_filter_on_asvar_ignored():
    assert (do_test('{% t "<xml>hello</xml> {var}" as text %}{{ text|safe }}',
                    #                                               ^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')
    assert (
        do_test('{% t "<xml>hello</xml> {var}" as text %}{{ text|escape }}',
                #                                               ^^^^^^^
                {'var': "<xml>world</xml>"},
                autoescape=True) ==
        '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;'
    )


def test_translate_variable():
    assert (do_test('{% ut source %}',
                    #      ^^^^^^
                    {'source': "<xml>hello</xml> {var}",
                     'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% ut source %}',
                    #      ^^^^^^
                    {'source': "<xml>hello</xml> {var}",
                     'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '<xml>hello</xml> <xml>world</xml>')
    assert (do_test('{% t source %}',
                    #     ^^^^^^
                    {'source': "<xml>hello</xml> {var}",
                     'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% t source %}',
                    #     ^^^^^^
                    {'source': "<xml>hello</xml> {var}",
                     'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '&lt;xml&gt;hello&lt;/xml&gt; <xml>world</xml>')


def test_filter_on_source_variable():
    assert (do_test('{% ut source|upper %}',
                    #            ^^^^^^
                    {'source': "<xml>hello</xml> {var}",
                     'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<XML>HELLO</XML> &LT;XML&GT;WORLD&LT;/XML&GT;')
    assert (do_test('{% ut source|upper %}',
                    #            ^^^^^^
                    {'source': "<xml>hello</xml> {var}",
                     'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '<XML>HELLO</XML> <XML>WORLD</XML>')
    assert (do_test('{% t source|upper %}',
                    #           ^^^^^^
                    {'source': "<xml>hello</xml> {var}",
                     'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&LT;XML&GT;HELLO&LT;/XML&GT; &LT;XML&GT;WORLD&LT;/XML&GT;')
    assert (do_test('{% t source|upper %}',
                    #           ^^^^^^
                    {'source': "<xml>hello</xml> {var}",
                     'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '&LT;XML&GT;HELLO&LT;/XML&GT; <XML>WORLD</XML>')


def test_safe_and_escape_filter_on_source_variable_ignored():
    assert (do_test('{% t source|safe %}',
                    #           ^^^^^
                    {'source': "<xml>hello</xml> {var}",
                     'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% t source|escape %}',
                    #           ^^^^^^^
                    {'source': "<xml>hello</xml> {var}",
                     'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')


def test_block():
    assert (do_test('{% t %}hello world{% endt %}') == 'hello world')
    #                       ^^^^^^^^^^^^^^^^^^^^^
    assert (do_test('{% t %}hello {var}{% endt %}', {'var': "world"}) ==
            #               ^^^^^^^^^^^^^^^^^^^^^
            'hello world')
    assert (do_test('{% ut %}<xml>hello</xml> {var}{% endut %}',
                    #       ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '<xml>hello</xml> <xml>world</xml>')
    assert (do_test('{% ut %}<xml>hello</xml> {var}{% endut %}',
                    #        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% t %}<xml>hello</xml> {var}{% endt %}',
                    #       ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=False) ==
            '&lt;xml&gt;hello&lt;/xml&gt; <xml>world</xml>')
    assert (do_test('{% t %}<xml>hello</xml> {var}{% endt %}',
                    #       ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')


def test_filter_on_block():
    assert (do_test('{% t |upper %}hello world{% endt %}') == 'HELLO WORLD')
    #                     ^^^^^^


def test_safe_and_escape_filter_on_block_ignored():
    assert (do_test('{% ut |safe %}<xml>hello</xml> {var}{% endut %}',
                    #      ^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% ut |escape %}<xml>hello</xml> {var}{% endut %}',
                    #      ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '<xml>hello</xml> &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% t |safe %}<xml>hello</xml> {var}{% endt %}',
                    #     ^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')
    assert (do_test('{% t |escape %}<xml>hello</xml> {var}{% endt %}',
                    #     ^^^^^^^
                    {'var': "<xml>world</xml>"},
                    autoescape=True) ==
            '&lt;xml&gt;hello&lt;/xml&gt; &lt;xml&gt;world&lt;/xml&gt;')


def test_translates_hashed_key():
    hello_key = generate_hashed_key(string='hello', context=None)
    tx._cache.update({'fr': (True, {hello_key: {'string': "bonjour"}})})
    assert do_test('{% t "hello" %}', lang_code="fr") == "bonjour"


def test_translates_source_key():
    hello_key = generate_key(string='hello', context=None)
    tx._cache.update({'fr': (True, {hello_key: {'string': "bonjour"}})})
    assert do_test('{% t "hello" %}', lang_code="fr") == "bonjour"


def test_translation_missing_hashed_key():
    old_missing_policy = tx._missing_policy
    tx._missing_policy = SourceStringPolicy()

    tx._cache._translations_by_lang = {}
    assert do_test('{% t "hello" %}', lang_code="fr") == "hello"

    hello_key = generate_hashed_key(string='hello', context=None)
    tx._cache.update({'fr': (True, {hello_key: {'string': None}})})
    assert do_test('{% t "hello" %}', lang_code="fr") == "hello"

    tx._missing_policy = old_missing_policy


def test_translation_missing_source_key():
    old_missing_policy = tx._missing_policy
    tx._missing_policy = SourceStringPolicy()

    tx._cache._translations_by_lang = {}
    assert do_test('{% t "hello" %}', lang_code="fr") == "hello"

    hello_key = generate_key(string='hello', context=None)
    tx._cache.update({'fr': (True, {hello_key: {'string': None}})})
    assert do_test('{% t "hello" %}', lang_code="fr") == "hello"

    tx._missing_policy = old_missing_policy


def test_escaping_is_done_on_translation():
    hello_key = generate_key(string='hello', context=None)
    tx._cache.update(
        {'fr': (True, {hello_key: {'string': "<xml>bonjour</xml>"}})})
    assert (do_test('{% t "hello" %}', lang_code="fr") ==
            '&lt;xml&gt;bonjour&lt;/xml&gt;')


def test_source_filter_is_applied_on_translation():
    # 'hello' => 'bonjour', 'HELLO' => 'I like pancakes'
    hello_key = generate_key(string='hello', context=None)
    HELLO_key = generate_key(string='HELLO', context=None)
    tx._cache.update(
        {'fr': (
            True,
            {
                hello_key: {'string': "bonjour"},
                HELLO_key: {'string': "I like pancakes"}
            }
        )})

    assert do_test('{% t "hello"|upper %}', lang_code="fr") == "BONJOUR"
    # If the filter was applied on the source string, we would get
    # 'I like pancakes'
    translation.activate('en-us')


def test_get_icu_keys():
    assert "username" in get_icu_keys("hello {username}")
    assert "cnt" in get_icu_keys("""
        {cnt, plural,
            one {you have one message}
            other {you have # new messages}}
    """)

    # Nesting
    assert "username" in get_icu_keys("""
        {gender, select,
            male {{username} is a boy}
            female {{username} is a girl}}
    """)

    # Return empty on error
    assert get_icu_keys("{{{") == set()



from __future__ import unicode_literals

import ast
import re
from collections import namedtuple

from transifex.common._compat import string_types
from transifex.common.utils import generate_key, make_hashable
from transifex.native import consts
from transifex.native.consts import KEY_CONTEXT, KEY_KEY

# PEP 263 magic comment for source code encodings
# e.g. "# -*- coding: <encoding name> -*-"

ENCODING_PATTERN = re.compile(r'#.*coding[:=]\s*utf-?8', re.IGNORECASE)


# A list of the translate functions that TxNative supports,
# in the syntax of module.deeper_module..function
DEFAULT_MODULES = [
    'transifex.native.translate',
    'transifex.native.lazy_translate',
]


Import = namedtuple('Import', ('module', 'function', 'node'))


class SourceString(object):
    """A data object that contains information about a source string."""

    def __init__(self, string, _context=None, fkeygen=generate_key, **meta):
        """Constructor.

        :param unicode string: the source string  # check ./consts.py
        :param unicode _context: an optional context that accompanies
            the source string
        :param func fkeygen: key generator function
        """
        custom_key = meta.get(KEY_KEY, None)
        self.key = (
            custom_key if custom_key
            else fkeygen(string=string, context=_context)
        )
        self.string = string
        self.context = (
            [x.strip() for x in _context.split(',')] if _context
            else None
        )  # type: list
        self.meta = self._transform_meta(meta)

    @property
    def occurrences(self):
        """An optional occurrence field, shows were the string is located

        :rtype: unicode
        """
        return self.meta.get(consts.KEY_OCCURRENCES, [])

    @occurrences.setter
    def occurrences(self, value):
        self.meta.setdefault(consts.KEY_OCCURRENCES, []).extend(value)

    @property
    def developer_comment(self):
        """An optional developer comment for the string.

        :rtype: unicode
        """
        return self.meta.get(consts.KEY_DEVELOPER_COMMENT)

    @property
    def character_limit(self):
        """An optional character limit for the string or None if not defined.

        :rtype: int
        """
        return self.meta.get(consts.KEY_CHARACTER_LIMIT)

    @property
    def tags(self):
        """A list of tags.

        :rtype: list
        """
        return self.meta.get(consts.KEY_TAGS, [])

    def _transform_meta(self, meta):
        """Transform values in meta object, whenever applicable.

        :param dict meta:
        :return: the same dictionary, with some values potentially altered
        :rtype: dict
        """
        tags = meta.get(consts.KEY_TAGS)
        if tags and isinstance(tags, string_types):
            meta[consts.KEY_TAGS] = [x.strip() for x in tags.split(',')]

        return {
            k: v for k, v in meta.items()
            if k in consts.ALL_KEYS
        }

    def __repr__(self):
        return u'<{}: {}>'.format(
            self.__class__.__name__,
            u' '.join((self.context or []) + [self.string]),
        )

    def __eq__(self, other):
        """Object equality.

        :param SourceString other: the instance to compare to
        :return: True if the instances are equal, False otherwise
        :rtype: bool
        """
        return hash(self) == hash(other)

    def __hash__(self):
        return hash((self.key, make_hashable(self.meta)))


class Extractor(object):
    """Extracts translatable source strings from Python files.

    It also keeps stats about any errors that have occurred.

    It allows clients to register custom module/function paths in order
    to support translate modules and functions that exist in framework
    implementations that use native, such as a Django or a Flask SDK.
    """

    def __init__(self):
        self.errors = []
        self._functions = []
        for path in DEFAULT_MODULES:
            self.register_functions(path)

    def register_functions(self, *func_paths):
        """Register a custom function to be detected during extraction.

        Each arg in `func_paths` must be a string, representing the full path
            of the function, like 'module.deeper_module.func_name'
        """
        for func_path in func_paths:
            nodes = func_path.split('.')
            if len(nodes) < 2:
                raise ValueError(
                    'Function path must contain at least a module and'
                    'a function, e.g. "my_module.translate"'
                )
            self._functions.append(
                ('.'.join(nodes[:-1]), nodes[-1])
            )

    def extract_strings(self, src, origin=None, fkeygen=generate_key):
        """Parse the given Python file string and extract translatable content.

        :param unicode src: a chunk of Python code
        :param str origin: the filename of the code, i.e. the filename
            it came from
        :return: a list of SourceString objects
        :rtype: list
        """
        # Replace utf-8 magic comment, to avoid getting a
        # "SyntaxError: encoding declaration in Unicode string"
        src = ENCODING_PATTERN.sub('# ', src)
        try:
            tree = ast.parse(src)
            visitor = CallDetectionVisitor(self._functions)
            visitor.visit(tree)
            source_strings, linenos = parse_source_strings(
                visitor.function_calls, fkeygen)
            # add file path to the string occurrence along with already
            # included number
        except Exception as e:
            # Store an exception for this particular file
            self.errors.append((origin, e))
            source_strings, linenos = [], []

        for src_str, lineno in zip(source_strings, linenos):
            src_str.occurrences = ["{}:{}".format(origin, lineno)]
        return source_strings


class CallDetectionVisitor(ast.NodeVisitor):
    """A visitor subclass that detects externally-defined imports and
    function methods and stores them.

    Subclasses can provide additional functionality that takes the imports
    and function calls and does something more useful with them.

    NOTE: in order for a function call to be detected, the corresponding
    `import` statement must appear before in the syntax tree. If we want
    to support calls that appear before imports, we could visit the tree
    twice, once to detect the imports and once to detect the calls
    and create the strings.
    """

    def __init__(self, registered_calls):
        """Constructor.

        Each item in `registered_calls` should be a tuple like:
          ('a.b.c', 'translate_func')
        which translates to calls like:
          from a.b.c import translate_func as tr; tr(...)
          from a.b import c; c.translate_func(...)
        and so on

        :param list registered_calls: a list of 2-tuples each with information
            on how to detect the imports and the calls
        """
        super(CallDetectionVisitor, self).__init__()

        # Contains a list of module/function pairs in dot notation
        # that are provided in the constructor. These are the imports
        # and function calls that the visitor will detect as being
        # "of interest".
        # Example: [('a.b', 'translate'), ('foo', 'trans')]
        self._registered_calls = registered_calls

        # Dynamically populated when parsing import statements,
        # this list will contain the module/function pairs that are
        # actually found in the current syntax tree (e.g. in a specific
        # Python file). This way, only calls that are actually made to
        # registered functions will be matched for each file, instead
        # of matching any function that has a name identical to what
        # a certain module provides (e.g. a `translate()` method of
        # another module).
        # Supports 'as' syntax, e.g. import translate as _trans
        # Contains `Import` objects.
        self.imports = []

        # This dictionary will contain all imports per Import/ImportFrom node.
        # It stores all imports, not only those that correspond to registered
        # calls.
        # For example, with the following statements:
        #   from a.b.c import gettext as _, e, d as _d
        # it will become:
        #   {
        #       <ImportFrom>: {
        #           'module': 'a.b.c',
        #           'imports': [('gettext', '_'), ('e', None), ('d', '_d')],
        #       },
        #   }
        self.imports_per_node = {}

        # For each detected function call that matches the detected imports
        # an ast.Call object will be stored here
        self.function_calls = []

        # We store each % operation here to process later, by its proceeding
        # function call node
        self.modulos = {}

    def visit_Import(self, node):
        """Support the 'import native as _native' type of imports.

        Given a supported function path that looks like this:
            >>> 'a.b.c.d.translate'
        takes an import statement that looks like this:
            >>> import a.b as _b
        and comes up with a tuple like this:
            >>> ('_b.c.d', 'translate')
        """
        self.generic_visit(node)

        for module_path, func_name in self._registered_calls:
            name = node.names[0].name
            as_name = node.names[0].asname
            # e.g. module_path='a.b.c', func_name='translate',
            #      name='a.b', asname='_c'
            if not module_path.startswith(name):
                continue

            remaining_module_path = module_path.replace(name, '')
            remaining_module_path = remaining_module_path.lstrip('.')

            if not as_name:
                as_name = name.split('.')[-1]

            remaining_module_path = (
                '{}.{}'.format(as_name, remaining_module_path).rstrip('.')
            )

            self.imports.append(
                Import(remaining_module_path, func_name, node)
            )

    def visit_ImportFrom(self, node):
        """Support the 'from native import native as _native' type of imports.

        Given a supported function path that looks like this:
            >>> 'a.b.c.d.translate'
        takes an import statement that looks like this:
            >>> from a.b import c as _c
        and comes up with a tuple like this:
            >>> ('_c.d', 'translate', <ImportFrom node>)
        """
        self.generic_visit(node)

        # Store the information for each part of the import
        imports_in_node = {'module': node.module, 'imports': []}
        for name_obj in node.names:
            name = name_obj.name
            as_name = name_obj.asname
            import_unit = (name, as_name)
            if import_unit not in imports_in_node['imports']:
                imports_in_node['imports'].append(import_unit)

        # Loop through all registered module/function calls,
        # e.g. transifex.native.django.t
        # and see if the current import matches any of them
        for registered_module_path, registered_func_name in self._registered_calls:
            # e.g. registered_module_path='transifex.native.django',
            #      registered_func_name='t'

            module = node.module
            if not registered_module_path.startswith(module):
                continue

            # Loop through all 'import' statements, e.g.
            # from m import a, b, c -> loop through [a, b, c] objects
            for name_obj in node.names:
                name = name_obj.name
                as_name = name_obj.asname

                try:
                    # If the full function call in the code is identical to the
                    # registered function name, e.g. it's `t('...')`
                    if name == registered_func_name:
                        registered_func_name = as_name or name
                        remaining_module_path = ''
                    else:
                        modules = registered_module_path.split('.')
                        if name in modules:
                            modules = modules[modules.index(name):]
                            if as_name and modules[0] == name:
                                modules[0] = as_name
                            remaining_module_path = '.'.join(modules)
                        else:
                            continue
                except Exception as e:
                    print(
                        'Error while visiting node: {}.{}{}: {}'.format(
                            module, name, (' as ' +
                                           as_name if as_name else ''),
                            e
                        )
                    )
                    continue

                # Store this import
                # Note that because self._registered_calls is a list,
                # we might append Import multiple objects with the same `node`
                self.imports.append(
                    Import(remaining_module_path, registered_func_name, node)
                )

        if imports_in_node:
            self.imports_per_node[node] = imports_in_node

    def visit_Call(self, node):
        """Extract a source string from a "translate" function call.

        Supports calls like:
          >>> translate('...')
          >>> a.b.c.translate('...')
        based on the supported calls that have been detected
        for the current syntax tree.
        """
        self.generic_visit(node)

        current_module_path, current_func_name = get_func_parts(node)

        # Check against all supported function calls and if there is a match
        # add the node for later processing
        for import_obj in self.imports:
            module_path = import_obj.module
            func_name = import_obj.function

            if module_path == current_module_path \
                    and func_name == current_func_name:
                self.function_calls.append(node)

    def visit_BinOp(self, node):
        """Store % operations to process later.

        We don't process them here because in _(...) % (...) statements
        the _(...) node hasn't been processed yet when the modulo operation
        is processed.
        """
        if isinstance(node.op, ast.Mod) and isinstance(node.left, ast.Call):
            func_call = node.left
            self.visit_Call(func_call)
            if self.function_calls and self.function_calls[-1] == func_call:
                self.modulos[func_call] = node
        else:
            self.generic_visit(node)


def parse_source_strings(nodes, fkeygen):
    """Parse the given function call nodes and return a list of SourceString
    objects.

    :param list nodes: a list of Node objects
    :param func fkeygen: key generator function

    :return:  a tuple of the source_strings along with the corresponding linenos
    :rtype: tuple
    """
    strings = []
    string_linenos = []
    for node in nodes:
        try:
            string = node.args[0].s
            # Context could be passed as an argument, e.g. t('str', 'context')
            context = node.args[1].s if len(node.args) > 1 else None
            # Find all custom parameters, e.g. developer comments etc
            params = {}
            for keyword in node.keywords:
                name, value = render_keyword(keyword)
                if value is not None:
                    params[name] = value
            string_linenos.append(node.lineno)
            # If no context was found before, maybe it was passed as a kwarg
            if context is None:
                context = params.pop(KEY_CONTEXT, None)

            strings.append(SourceString(string, context, fkeygen, **params))
        except Exception as e:
            raise AttributeError(
                'Invalid module/function format on line {} col {}: {}'.format(
                    node.lineno, node.col_offset, e
                )
            )

    return strings, string_linenos


def render_keyword(keyword):
    """Render the given keyword to a proper value.

    Processes Keyword objects and returns the keyword name along with
    a value of their corresponding type, i.e. a string, a number
    or a boolean.

    :param Keyword keyword: the keyword object
    :return: a tuple like (<key>, <value>)
    :rtype: tuple
    """
    if isinstance(keyword.value, ast.Str):
        val = keyword.value.s
    elif isinstance(keyword.value, ast.Num):
        val = keyword.value.n
    elif (isinstance(keyword.value, ast.Name)
          and keyword.value.id in ('True', 'False')):
        val = keyword.value.id == 'True'
    else:
        val = None

    return keyword.arg, val


def get_func_parts(node):
    # Find the full module/function path of the current calling node,
    # e.g. 'a.b.c.translate'.
    # The way to retrieve the module name and function name
    # differs a lot depending on the level of nesting, so try...catch blocks
    # are used in order to cover all cases
    modules = []
    try:
        current_node = node.func.value
        current_func_name = node.func.attr
        # Module nesting can be indefinite, so we need to follow it
        # to the end
        while True:
            try:
                modules.insert(0, current_node.attr)
                current_node = current_node.value
            except AttributeError:
                modules.insert(0, current_node.id)
                break
    except AttributeError:
        try:
            current_func_name = node.func.id
        except AttributeError:
            try:
                current_func_name = node.func.attr
            except AttributeError as e:
                raise AttributeError(
                    'Invalid module/function format on line {} '
                    'col {}: {}'.format(node.lineno, node.col_offset, e)
                )

    current_module_path = '.'.join(modules)

    return current_module_path, current_func_name



from __future__ import unicode_literals

from django.template.base import Lexer, Parser
from django.utils.encoding import force_text
from transifex.common._compat import string_types
from transifex.common.utils import generate_key
from transifex.native.consts import KEY_CONTEXT
from transifex.native.django.compat import TOKEN_BLOCK
from transifex.native.django.templatetags.transifex import do_t
from transifex.native.parsing import SourceString

# Django template consts
LOAD_TAG = 'load'
WITH_TAG = 'with'
ENDWITH_TAG = 'endwith'
COMMENT_TAG = 'comment'
ENDCOMMENT_TAG = 'endcomment'
TRANSLATE_TAGS = ['trans', 'translate']
BLOCK_TRANSLATE_TAGS = ['blocktrans', 'blocktranslate']
ENDBLOCK_TRANSLATE_TAGS = ['endblocktrans', 'endblocktranslate']
DJANGO_i18n_TAG_NAME = 'i18n'
TRANSIFEX_TAG_NAME = 'transifex'
COMMENT_FOUND = object()
COPY_AS_IS = object()


# hack to make the identity function
# signature compatible to all template filters
def identity(obj, var1=None, var2=None, var3=None, var4=None, var5=None):  # pragma no cover
    return obj  # pragma no cover


def find_filter_identity(filter_name):
    """Return a filter that does no filtering, i.e always returns all items.

    This function supposedly returns a filter object, which in this case
    does nothing other than returning the variable as is.

    :param str filter_name:
    :return: a callable that represents a filter that does not filtering
    :rtype: callable
    """
    return identity


def tnode_to_source_string(tnode, fkeygen):
    """ Convert a parsed TNode to a SourceString instance.

        TNode is what the template tag implementation will return having
        processed the contents from the template. A SourceString is a data
        object which exposes information in a useful way for pushing to
        transifex.

        :param func fkeygen: key generator function
    """
    if not isinstance(tnode.source_string.var, string_types):
        return None
    meta = {}
    for key, value in tnode.params.items():
        if len(value.filters) != 0:
            continue
        if isinstance(value.var, string_types):
            meta[key] = value.var
        elif getattr(value.var, 'literal', None) is not None:
            meta[key] = value.var.literal
    _context = meta.pop(KEY_CONTEXT, None)

    return SourceString(
        tnode.source_string.var, _context, fkeygen, **meta)


def extract_transifex_template_strings(
        src, origin=None, charset='utf-8', fkeygen=generate_key):
    """Parse the given template and extract translatable content
    based on the syntax supported by Transifex Native.

    Supports the {% t %} and {% ut %} template tags.

    :param unicode src: the whole Django template
    :param str origin: an optional context for the filename of the source,
        e.g. the file name
    :param str charset: the character set to use
    :param func fkeygen: key generator function
    :return: a list of SourceString objects
    :rtype: list
    """
    src = force_text(src, charset)
    tokens = Lexer(src).tokenize()
    parser = Parser(tokens, {}, [], origin)
    # Since no template libraries are loaded when this code is running,
    # we need to override the find function in order to use the functionality
    # of the Parser class. The overridden function returns the object as given.
    # Without the override, a KeyError is raised inside the parser.
    parser.find_filter = find_filter_identity

    strings = []
    while parser.tokens:
        token = parser.next_token()
        if (token.token_type == TOKEN_BLOCK and
                token.split_contents()[0] in ('t', 'ut')):
            tnode = do_t(parser, token)
            source_string = tnode_to_source_string(tnode, fkeygen)
            if source_string is None:
                continue
            if token.lineno and origin:
                source_string.occurrences = [
                    "{}:{}".format(origin, token.lineno)]

            strings.append(source_string)

    return strings
